# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import logging
import os
import shutil
import stat
import tempfile
import unittest
from zoneinfo import ZoneInfo


from dulwich.objects import Blob, Commit, Tree
from dulwich.repo import Repo

from xandikos.store import (
    DuplicateUidError,
    File,
    Filter,
    InvalidETag,
    NoSuchItem,
    Store,
    start_eager_indexing,
)

from xandikos.icalendar import ICalendarFile, CalendarFilter
from xandikos.store.git import BareGitStore, GitStore, TreeGitStore
from xandikos.store.memory import MemoryStore
from xandikos.store.vdir import VdirStore

EXAMPLE_VCALENDAR1 = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20150314T223512Z
DTSTAMP:20150527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something
UID:bdc22720-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""

EXAMPLE_VCALENDAR1_NORMALIZED = b"""\
BEGIN:VCALENDAR\r
VERSION:2.0\r
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN\r
BEGIN:VTODO\r
CREATED:20150314T223512Z\r
DTSTAMP:20150527T221952Z\r
LAST-MODIFIED:20150314T223512Z\r
STATUS:NEEDS-ACTION\r
SUMMARY:do something\r
UID:bdc22720-b9e1-42c9-89c2-a85405d8fbff\r
END:VTODO\r
END:VCALENDAR\r
"""

EXAMPLE_VCALENDAR2 = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20120314T223512Z
DTSTAMP:20130527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something else
UID:bdc22764-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""

EXAMPLE_VCALENDAR2_NORMALIZED = b"""\
BEGIN:VCALENDAR\r
VERSION:2.0\r
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN\r
BEGIN:VTODO\r
CREATED:20120314T223512Z\r
DTSTAMP:20130527T221952Z\r
LAST-MODIFIED:20150314T223512Z\r
STATUS:NEEDS-ACTION\r
SUMMARY:do something else\r
UID:bdc22764-b9e1-42c9-89c2-a85405d8fbff\r
END:VTODO\r
END:VCALENDAR\r
"""

EXAMPLE_VCALENDAR_NO_UID = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20120314T223512Z
DTSTAMP:20130527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something without uid
END:VTODO
END:VCALENDAR
"""


class BaseStoreTest:
    def test_import_one(self):
        gc = self.create_store()
        (name, etag) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertIsInstance(etag, str)
        self.assertEqual(
            [("foo.ics", "text/calendar", etag)], list(gc.iter_with_etag())
        )

    def test_with_filter(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])

        class DummyFilter(Filter):
            content_type = "text/calendar"

            def __init__(self, text) -> None:
                self.text = text

            def check(self, name, resource):
                return self.text in b"".join(resource.content)

        self.assertEqual(
            2, len(list(gc.iter_with_filter(filter=DummyFilter(b"do something"))))
        )

        [(ret_name, ret_file, ret_etag)] = list(
            gc.iter_with_filter(filter=DummyFilter(b"do something else"))
        )
        self.assertEqual(ret_name, name2)
        self.assertEqual(ret_etag, etag2)
        self.assertEqual(ret_file.content_type, "text/calendar")
        self.assertEqual(
            b"".join(ret_file.content),
            EXAMPLE_VCALENDAR2.replace(b"\n", b"\r\n"),
        )

    def test_get_by_index(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])
        (name3, etag3) = gc.import_one(
            "bar.txt", "text/plain", [b"Not a calendar file."]
        )
        self.assertEqual({}, dict(gc.index_manager.desired))

        filtertext = "C=VCALENDAR/C=VTODO/P=SUMMARY"

        class DummyFilter(Filter):
            content_type = "text/calendar"

            def __init__(self, text) -> None:
                self.text = text

            def index_keys(self):
                return [[filtertext]]

            def check_from_indexes(self, name, index_values):
                return any(self.text in v for v in index_values[filtertext])

            def check(self, name, resource):
                return self.text in b"".join(resource.content)

        self.assertEqual(
            2, len(list(gc.iter_with_filter(filter=DummyFilter(b"do something"))))
        )

        [(ret_name, ret_file, ret_etag)] = list(
            gc.iter_with_filter(filter=DummyFilter(b"do something else"))
        )
        self.assertEqual({filtertext: 2}, dict(gc.index_manager.desired))

        # Force index
        gc.index.reset([filtertext])

        [(ret_name, ret_file, ret_etag)] = list(
            gc.iter_with_filter(filter=DummyFilter(b"do something else"))
        )
        self.assertEqual({filtertext: 2}, dict(gc.index_manager.desired))

        self.assertEqual(ret_name, name2)
        self.assertEqual(ret_etag, etag2)
        self.assertEqual(ret_file.content_type, "text/calendar")
        self.assertEqual(
            b"".join(ret_file.content),
            EXAMPLE_VCALENDAR2.replace(b"\n", b"\r\n"),
        )

    def test_import_one_duplicate_uid(self):
        gc = self.create_store()
        (name, etag) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertRaises(
            DuplicateUidError,
            gc.import_one,
            "bar.ics",
            "text/calendar",
            [EXAMPLE_VCALENDAR1],
        )

    def test_import_one_duplicate_name(self):
        gc = self.create_store()
        (name, etag) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name, etag) = gc.import_one(
            "foo.ics", "text/calendar", [EXAMPLE_VCALENDAR2], replace_etag=etag
        )
        (name, etag) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertRaises(
            InvalidETag,
            gc.import_one,
            "foo.ics",
            "text/calendar",
            [EXAMPLE_VCALENDAR2],
            replace_etag="invalidetag",
        )

    def test_get_raw(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])
        self.assertEqual(
            EXAMPLE_VCALENDAR1_NORMALIZED,
            b"".join(gc._get_raw("foo.ics", etag1)),
        )
        self.assertEqual(
            EXAMPLE_VCALENDAR2_NORMALIZED,
            b"".join(gc._get_raw("bar.ics", etag2)),
        )
        self.assertRaises(KeyError, gc._get_raw, "missing.ics", "01" * 20)

    def test_get_file(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name1, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])
        f1 = gc.get_file("foo.ics", "text/calendar", etag1)
        self.assertEqual(EXAMPLE_VCALENDAR1_NORMALIZED, b"".join(f1.content))
        self.assertEqual("text/calendar", f1.content_type)
        f2 = gc.get_file("bar.ics", "text/calendar", etag2)
        self.assertEqual(EXAMPLE_VCALENDAR2_NORMALIZED, b"".join(f2.content))
        self.assertEqual("text/calendar", f2.content_type)
        self.assertRaises(KeyError, gc._get_raw, "missing.ics", "01" * 20)

    def test_delete_one(self):
        gc = self.create_store()
        self.assertEqual([], list(gc.iter_with_etag()))
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertEqual(
            [("foo.ics", "text/calendar", etag1)], list(gc.iter_with_etag())
        )
        gc.delete_one("foo.ics")
        self.assertEqual([], list(gc.iter_with_etag()))

    def test_delete_one_with_etag(self):
        gc = self.create_store()
        self.assertEqual([], list(gc.iter_with_etag()))
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertEqual(
            [("foo.ics", "text/calendar", etag1)], list(gc.iter_with_etag())
        )
        gc.delete_one("foo.ics", etag=etag1)
        self.assertEqual([], list(gc.iter_with_etag()))

    def test_delete_one_nonexistant(self):
        gc = self.create_store()
        self.assertRaises(NoSuchItem, gc.delete_one, "foo.ics")

    def test_delete_one_invalid_etag(self):
        gc = self.create_store()
        self.assertEqual([], list(gc.iter_with_etag()))
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])
        self.assertEqual(
            {
                ("foo.ics", "text/calendar", etag1),
                ("bar.ics", "text/calendar", etag2),
            },
            set(gc.iter_with_etag()),
        )
        self.assertRaises(InvalidETag, gc.delete_one, "foo.ics", etag=etag2)
        self.assertEqual(
            {
                ("foo.ics", "text/calendar", etag1),
                ("bar.ics", "text/calendar", etag2),
            },
            set(gc.iter_with_etag()),
        )


class VdirStoreTest(BaseStoreTest, unittest.TestCase):
    kls = VdirStore

    def create_store(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        store = self.kls.create(os.path.join(d, "store"))
        store.load_extra_file_handler(ICalendarFile)
        return store


class MemoryStoreTest(BaseStoreTest, unittest.TestCase):
    kls = MemoryStore

    def create_store(self):
        store = self.kls()
        store.load_extra_file_handler(ICalendarFile)
        return store


class BaseGitStoreTest(BaseStoreTest):
    kls: type[Store]

    def create_store(self):
        raise NotImplementedError(self.create_store)

    def add_blob(self, gc, name, contents):
        raise NotImplementedError(self.add_blob)

    def test_create(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        gc = self.kls.create(os.path.join(d, "store"))
        self.assertIsInstance(gc, GitStore)
        self.assertEqual(gc.repo.path, os.path.join(d, "store"))

    def test_iter_with_etag_missing_uid(self):
        logging.getLogger("").setLevel(logging.ERROR)
        gc = self.create_store()
        bid = self.add_blob(gc, "foo.ics", EXAMPLE_VCALENDAR_NO_UID)
        self.assertEqual([("foo.ics", "text/calendar", bid)], list(gc.iter_with_etag()))
        gc._scan_uids()
        logging.getLogger("").setLevel(logging.NOTSET)

    def test_iter_with_etag(self):
        gc = self.create_store()
        bid = self.add_blob(gc, "foo.ics", EXAMPLE_VCALENDAR1)
        self.assertEqual([("foo.ics", "text/calendar", bid)], list(gc.iter_with_etag()))

    def test_get_description_from_git_config(self):
        gc = self.create_store()
        config = gc.repo.get_config()
        config.set(b"xandikos", b"test", b"test")
        if getattr(config, "path", None):
            config.write_to_path()
        gc.repo.set_description(b"a repo description")
        self.assertEqual(gc.get_description(), "a repo description")

    def test_displayname(self):
        gc = self.create_store()
        self.assertIs(None, gc.get_color())
        c = gc.repo.get_config()
        c.set(b"xandikos", b"displayname", b"a name")
        if getattr(c, "path", None):
            c.write_to_path()
        self.assertEqual("a name", gc.get_displayname())

    def test_get_color(self):
        gc = self.create_store()
        self.assertIs(None, gc.get_color())
        c = gc.repo.get_config()
        c.set(b"xandikos", b"color", b"334433")
        if getattr(c, "path", None):
            c.write_to_path()
        self.assertEqual("334433", gc.get_color())

    def test_get_source_url(self):
        gc = self.create_store()
        self.assertIs(None, gc.get_source_url())
        c = gc.repo.get_config()
        c.set(b"xandikos", b"source", b"www.google.com")
        if getattr(c, "path", None):
            c.write_to_path()
        self.assertEqual("www.google.com", gc.get_source_url())

    def test_default_no_subdirectories(self):
        gc = self.create_store()
        self.assertEqual([], gc.subdirectories())

    def _get_last_commit_message(self, gc):
        walker = gc.repo.get_walker(include=[gc.repo.refs[gc.ref]])
        entry = next(iter(walker))
        return entry.commit.message.decode("utf-8")

    def test_import_remote_user_in_commit_message(self):
        gc = self.create_store()
        gc.import_one(
            "foo.ics",
            "text/calendar",
            [EXAMPLE_VCALENDAR1],
            remote_user="alice",
        )
        message = self._get_last_commit_message(gc)
        self.assertIn("Remote-User: alice", message)

    def test_import_no_remote_user_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        message = self._get_last_commit_message(gc)
        self.assertNotIn("Remote-User:", message)

    def test_delete_remote_user_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        gc.delete_one("foo.ics", remote_user="alice")
        message = self._get_last_commit_message(gc)
        self.assertIn("Remote-User: alice", message)

    def test_delete_no_remote_user_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        gc.delete_one("foo.ics")
        message = self._get_last_commit_message(gc)
        self.assertNotIn("Remote-User:", message)

    def test_import_requester_in_commit_message(self):
        gc = self.create_store()
        gc.import_one(
            "foo.ics",
            "text/calendar",
            [EXAMPLE_VCALENDAR1],
            requester="iOS/26.3.1 remindd/3900.1",
        )
        message = self._get_last_commit_message(gc)
        self.assertIn("Requester: iOS/26.3.1 remindd/3900.1", message)

    def test_import_no_requester_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        message = self._get_last_commit_message(gc)
        self.assertNotIn("Requester:", message)

    def test_delete_requester_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        gc.delete_one("foo.ics", requester="iOS/26.3.1 remindd/3900.1")
        message = self._get_last_commit_message(gc)
        self.assertIn("Requester: iOS/26.3.1 remindd/3900.1", message)

    def test_delete_no_requester_in_commit_message(self):
        gc = self.create_store()
        gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        gc.delete_one("foo.ics")
        message = self._get_last_commit_message(gc)
        self.assertNotIn("Requester:", message)

    def test_import_only_once(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        self.assertEqual(name1, name2)
        self.assertEqual(etag1, etag2)
        walker = gc.repo.get_walker(include=[gc.repo.refs[gc.ref]])
        self.assertEqual(1, len([w.commit for w in walker]))


class GitStoreTest(unittest.TestCase):
    def test_open_from_path_bare(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        Repo.init_bare(d)
        gc = GitStore.open_from_path(d)
        self.assertIsInstance(gc, BareGitStore)
        self.assertEqual(gc.repo.path, d)

    def test_open_from_path_tree(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        Repo.init(d)
        gc = GitStore.open_from_path(d)
        self.assertIsInstance(gc, TreeGitStore)
        self.assertEqual(gc.repo.path, d)


class BareGitStoreTest(BaseGitStoreTest, unittest.TestCase):
    kls = BareGitStore

    def create_store(self):
        store = BareGitStore.create_memory()
        store.load_extra_file_handler(ICalendarFile)
        return store

    def test_create_memory(self):
        gc = BareGitStore.create_memory()
        self.assertIsInstance(gc, GitStore)

    def add_blob(self, gc, name, contents):
        b = Blob.from_string(contents)
        t = Tree()
        t.add(name.encode("utf-8"), 0o644 | stat.S_IFREG, b.id)
        c = Commit()
        c.tree = t.id
        c.committer = c.author = b"Somebody <foo@example.com>"
        c.commit_time = c.author_time = 800000
        c.commit_timezone = c.author_timezone = 0
        c.message = b"do something"
        gc.repo.object_store.add_objects([(b, None), (t, None), (c, None)])
        gc.repo[gc.ref] = c.id
        return b.id.decode("ascii")

    def test_get_ctag(self):
        gc = self.create_store()
        self.assertEqual(Tree().id.decode("ascii"), gc.get_ctag())
        self.add_blob(gc, "foo.ics", EXAMPLE_VCALENDAR1)
        self.assertEqual(gc._get_current_tree().id.decode("ascii"), gc.get_ctag())

    def test_tree_caching(self):
        """Test that the current tree is cached and not re-read on every call."""
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])

        # First call populates the cache
        self.assertEqual(etag1, gc.get_etag("foo.ics"))

        # Subsequent calls should return the same tree object
        tree1 = gc._get_current_tree()
        tree2 = gc._get_current_tree()
        self.assertIs(tree1, tree2)

        # After a write, cache should be invalidated (ref changes)
        gc.delete_one("bar.ics")
        tree3 = gc._get_current_tree()
        self.assertIsNot(tree1, tree3)

        # get_etag still works for remaining item
        self.assertEqual(etag1, gc.get_etag("foo.ics"))
        self.assertRaises(KeyError, gc.get_etag, "bar.ics")


class TreeGitStoreTest(BaseGitStoreTest, unittest.TestCase):
    kls = TreeGitStore

    def create_store(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        store = self.kls.create(os.path.join(d, "store"))
        store.load_extra_file_handler(ICalendarFile)
        return store

    def add_blob(self, gc, name, contents):
        with open(os.path.join(gc.repo.path, name), "wb") as f:
            f.write(contents)
        gc.repo.get_worktree().stage(name.encode("utf-8"))
        return Blob.from_string(contents).id.decode("ascii")

    def test_index_caching(self):
        """Test that the git index is cached and not re-read on every call."""
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])
        (name2, etag2) = gc.import_one("bar.ics", "text/calendar", [EXAMPLE_VCALENDAR2])

        # First call populates the cache
        self.assertEqual(etag1, gc.get_etag("foo.ics"))

        # Subsequent calls should use cached index (same object)
        idx1, ctag1 = gc._open_index()
        idx2, ctag2 = gc._open_index()
        self.assertIs(idx1, idx2)

        # After a write, cache should be invalidated
        gc.delete_one("bar.ics")
        idx3, _ctag3 = gc._open_index()
        self.assertIsNot(idx1, idx3)

        # get_etag still works for remaining item
        self.assertEqual(etag1, gc.get_etag("foo.ics"))
        self.assertRaises(KeyError, gc.get_etag, "bar.ics")

    def test_index_cache_invalidated_on_delete(self):
        gc = self.create_store()
        (name1, etag1) = gc.import_one("foo.ics", "text/calendar", [EXAMPLE_VCALENDAR1])

        idx_before, _ctag_before = gc._open_index()
        gc.delete_one("foo.ics")
        idx_after, _ctag_after = gc._open_index()
        self.assertIsNot(idx_before, idx_after)
        self.assertRaises(KeyError, gc.get_etag, "foo.ics")


class ExtractRegularUIDTests(unittest.TestCase):
    def test_extract_no_uid(self):
        fi = File([EXAMPLE_VCALENDAR_NO_UID], "text/bla")
        self.assertRaises(NotImplementedError, fi.get_uid)


class ParanoidModeTests(unittest.TestCase):
    """Test for the original issue #235: AssertionError in paranoid mode with index_threshold=0."""

    def test_index_keys_return_type_bug(self):
        """Test that demonstrates the index_keys() return type bug directly."""
        # Create a filter with multiple components/properties to trigger the extend() bug
        filter = CalendarFilter(ZoneInfo("UTC"))
        component_filter = filter.filter_subcomponent("VCALENDAR")
        todo_filter = component_filter.filter_subcomponent("VTODO")
        todo_filter.filter_property("SUMMARY")
        todo_filter.filter_property("CREATED")

        index_keys = filter.index_keys()

        # The fixed code should return a list of lists (AND-list of OR-options)
        # Each inner list represents OR-options for a single AND requirement
        self.assertIsInstance(index_keys, list)
        if index_keys:
            # Each element should be a list of strings (OR-options)
            for key_group in index_keys:
                self.assertIsInstance(
                    key_group,
                    list,
                    f"Expected list of lists, but got list containing {type(key_group)} in {index_keys}",
                )
                for key in key_group:
                    self.assertIsInstance(
                        key,
                        str,
                        f"Expected strings in inner lists, but got {type(key)} in {key_group}",
                    )

        # Expected structure with the fix: [['C=VCALENDAR/C=VTODO/P=SUMMARY'], ['C=VCALENDAR/C=VTODO/P=CREATED']]
        expected_keys = [
            ["C=VCALENDAR/C=VTODO/P=SUMMARY"],
            ["C=VCALENDAR/C=VTODO/P=CREATED"],
        ]
        self.assertEqual(index_keys, expected_keys)

    def test_paranoid_mode_index_threshold_zero(self):
        """Test that reproduces the original AssertionError when using paranoid mode with index_threshold=0.

        This test verifies the fix for: "AssertionError: index based filter not matching real file filter"
        which occurred when double_check_indexes=True and index_threshold=0.
        """
        # Create a temporary store with paranoid mode enabled and index_threshold=0
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)

        store_path = os.path.join(d, "store")
        os.mkdir(store_path)
        repo = Repo.init_bare(store_path)
        repo._autogc_disabled = True

        store = BareGitStore(
            repo,
            double_check_indexes=True,  # Enable paranoid mode
            index_threshold=0,  # Force immediate indexing
        )
        store.load_extra_file_handler(ICalendarFile)

        # Ensure repository is properly closed after test
        self.addCleanup(repo.close)

        # Import a calendar file
        (name, etag) = store.import_one(
            "test.ics", "text/calendar", [EXAMPLE_VCALENDAR1]
        )

        # Create a CalDAV filter that would trigger the bug
        # The old bug happened when index_keys() returned list[str] instead of list[list[str]]
        # causing find_present_keys to iterate over characters instead of keys
        filter = CalendarFilter(ZoneInfo("UTC"))
        component_filter = filter.filter_subcomponent("VCALENDAR")
        todo_filter = component_filter.filter_subcomponent("VTODO")
        todo_filter.filter_property("SUMMARY")

        # Force a scenario that uses the index by pre-indexing and then filtering
        # First, let the filter create some index entries
        results = list(store.iter_with_filter(filter))

        # The old code would sometimes work on first run but fail on subsequent runs
        # when the index state differed. Let's try again to trigger the comparison failure.
        results2 = list(store.iter_with_filter(filter))

        # Verify we get the expected results
        self.assertEqual(len(results), 1)
        self.assertEqual(len(results2), 1)
        result_name, result_file, result_etag = results[0]
        self.assertEqual(result_name, name)
        self.assertEqual(result_etag, etag)


class EagerIndexingTest(unittest.TestCase):
    def _create_bare_store(self):
        d = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, d)
        store_path = os.path.join(d, "store")
        os.mkdir(store_path)
        repo = Repo.init_bare(store_path)
        repo._autogc_disabled = True
        self.addCleanup(repo.close)
        store = BareGitStore(repo)
        store.load_extra_file_handler(ICalendarFile)
        return store

    def test_start_eager_indexing_populates_index(self):
        store = self._create_bare_store()
        store.import_one("test.ics", "text/calendar", [EXAMPLE_VCALENDAR1])

        thread = start_eager_indexing(store)
        self.assertIsNotNone(thread)
        thread.join(timeout=10)
        self.assertFalse(thread.is_alive())

        # Index should now have the default keys
        available = set(store.index.available_keys())
        for key in ICalendarFile.default_index_keys():
            self.assertIn(key, available)

        # Index should contain the imported item's etag
        indexed_etags = list(store.index.iter_etags())
        self.assertEqual(len(indexed_etags), 1)

    def test_eager_indexing_filter_uses_index(self):
        store = self._create_bare_store()
        (name, etag) = store.import_one(
            "test.ics", "text/calendar", [EXAMPLE_VCALENDAR1]
        )

        thread = start_eager_indexing(store)
        thread.join(timeout=10)

        # Query with a filter that uses the indexed keys
        filter = CalendarFilter(ZoneInfo("UTC"))
        component_filter = filter.filter_subcomponent("VCALENDAR")
        component_filter.filter_subcomponent("VTODO")

        results = list(store.iter_with_filter(filter))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0][0], name)
        self.assertEqual(results[0][2], etag)

    def test_start_eager_indexing_no_handlers(self):
        store = self._create_bare_store()
        # Remove all file handlers so there are no default keys
        store.extra_file_handlers.clear()
        thread = start_eager_indexing(store)
        self.assertIsNone(thread)

    def test_default_index_keys(self):
        self.assertEqual(File.default_index_keys(), [])
        self.assertGreater(len(ICalendarFile.default_index_keys()), 0)
        self.assertIn(
            "C=VCALENDAR/C=VEVENT/P=DTSTART", ICalendarFile.default_index_keys()
        )

    def test_eager_indexing_skips_invalid_files(self):
        """Verify that invalid files are skipped with a warning, not crashing."""
        store = self._create_bare_store()
        # Import a valid file
        store.import_one("good.ics", "text/calendar", [EXAMPLE_VCALENDAR1])

        thread = start_eager_indexing(store)
        thread.join(timeout=10)
        self.assertFalse(thread.is_alive())

        # The valid file should be indexed
        indexed_etags = list(store.index.iter_etags())
        self.assertEqual(len(indexed_etags), 1)



# Xandikos
# Copyright (C) 2025 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Comprehensive filter combination tests for CardDAV addressbook-query (RFC 6352 §10.5)."""

import unittest

from xandikos.vcard import CardDAVFilter, VCardFile


class CardDAVFilterCombinationTests(unittest.TestCase):
    """Test various combinations of CardDAV filters (RFC 6352 Section 10.5)."""

    def setUp(self):
        """Set up test vCard data."""
        # Test contact 1: Alice with email and phone
        self.vcard1 = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Alice Smith\r\n",
                b"N:Smith;Alice;;;\r\n",
                b"EMAIL:alice@example.com\r\n",
                b"TEL;TYPE=WORK:+1-555-1234\r\n",
                b"ORG:Example Corp\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Test contact 2: Bob without email
        self.vcard2 = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Bob Jones\r\n",
                b"N:Jones;Bob;;;\r\n",
                b"TEL;TYPE=HOME:+1-555-5678\r\n",
                b"ORG:Acme Inc\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Test contact 3: Charlie with multiple properties
        self.vcard3 = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Charlie Brown\r\n",
                b"N:Brown;Charlie;;;\r\n",
                b"EMAIL:charlie@example.org\r\n",
                b"TEL;TYPE=WORK:+1-555-9999\r\n",
                b"TEL;TYPE=CELL:+1-555-0000\r\n",
                b"TITLE:Developer\r\n",
                b"ORG:Tech Solutions\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

    def test_prop_filter_fn_contains(self):
        """Test prop-filter with text-match contains.

        RFC 6352 Section 10.5.1: Text match with contains.
        """
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "Smith", collation="i;unicode-casemap", match_type="contains"
        )

        # vcard1 has "Alice Smith"
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2 has "Bob Jones" - no match
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_prop_filter_fn_equals(self):
        """Test prop-filter with text-match equals."""
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "Bob Jones", collation="i;unicode-casemap", match_type="equals"
        )

        self.assertFalse(filter_obj.check("vcard1.vcf", self.vcard1))
        self.assertTrue(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_prop_filter_fn_starts_with(self):
        """Test prop-filter with text-match starts-with."""
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "Alice", collation="i;unicode-casemap", match_type="starts-with"
        )

        # vcard1: "Alice Smith" starts with "Alice"
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: "Bob Jones" doesn't start with "Alice"
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_prop_filter_fn_ends_with(self):
        """Test prop-filter with text-match ends-with."""
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "Brown", collation="i;unicode-casemap", match_type="ends-with"
        )

        # vcard3: "Charlie Brown" ends with "Brown"
        self.assertTrue(filter_obj.check("vcard3.vcf", self.vcard3))
        # vcard1: "Alice Smith" doesn't end with "Brown"
        self.assertFalse(filter_obj.check("vcard1.vcf", self.vcard1))

    def test_prop_filter_is_not_defined(self):
        """Test prop-filter with is-not-defined.

        RFC 6352 Section 10.5.1: Property must not exist.
        """
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("EMAIL", is_not_defined=True)

        # vcard1 has EMAIL
        self.assertFalse(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2 has no EMAIL
        self.assertTrue(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_prop_filter_is_defined(self):
        """Test prop-filter without is-not-defined (property must exist)."""
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("EMAIL")

        # vcard1 has EMAIL
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2 has no EMAIL
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_multiple_prop_filters_anyof(self):
        """Test multiple prop-filters with anyof test (OR logic).

        RFC 6352 Section 10.5: Default test is anyof.
        """
        filter_obj = CardDAVFilter()
        # anyof is the default
        filter_obj.add_property_filter("FN").add_text_match(
            "Alice", match_type="contains"
        )
        filter_obj.add_property_filter("FN").add_text_match(
            "Charlie", match_type="contains"
        )

        # vcard1: has "Alice"
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: has neither "Alice" nor "Charlie"
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))
        # vcard3: has "Charlie"
        self.assertTrue(filter_obj.check("vcard3.vcf", self.vcard3))

    def test_multiple_prop_filters_allof(self):
        """Test multiple prop-filters with allof test (AND logic).

        RFC 6352 Section 10.5: Test attribute determines AND/OR.
        """
        filter_obj = CardDAVFilter()
        filter_obj.test = all  # allof (AND logic)
        filter_obj.add_property_filter("EMAIL")
        filter_obj.add_property_filter("ORG").add_text_match(
            "Example", match_type="contains"
        )

        # vcard1: has EMAIL and "Example Corp"
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: no EMAIL
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))
        # vcard3: has EMAIL but "Tech Solutions" not "Example"
        self.assertFalse(filter_obj.check("vcard3.vcf", self.vcard3))

    def test_param_filter_type(self):
        """Test param-filter for TYPE parameter.

        RFC 6352 Section 10.5.2: Parameter filtering.
        """
        filter_obj = CardDAVFilter()
        prop_filter = filter_obj.add_property_filter("TEL")
        prop_filter.add_param_filter("TYPE").add_text_match(
            "WORK", match_type="contains"
        )

        # vcard1: has TEL with TYPE=WORK
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: has TEL with TYPE=HOME
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))
        # vcard3: has TEL with TYPE=WORK
        self.assertTrue(filter_obj.check("vcard3.vcf", self.vcard3))

    def test_param_filter_is_not_defined(self):
        """Test param-filter with is-not-defined."""
        filter_obj = CardDAVFilter()
        prop_filter = filter_obj.add_property_filter("TEL")
        prop_filter.add_param_filter("TYPE", is_not_defined=True)

        # All test vcards have TEL with TYPE parameter
        self.assertFalse(filter_obj.check("vcard1.vcf", self.vcard1))
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))
        self.assertFalse(filter_obj.check("vcard3.vcf", self.vcard3))

    def test_text_match_case_insensitive(self):
        """Test text-match with case insensitivity.

        RFC 6352 Section 10.5.1: Collation determines case sensitivity.
        """
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "SMITH", collation="i;unicode-casemap", match_type="contains"
        )

        # vcard1 has "Alice Smith" - should match with case-insensitive
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))

    def test_text_match_negate_condition(self):
        """Test text-match with negate-condition.

        RFC 6352 Section 10.5.1: Negate-condition inverts the match.
        """
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "Alice", match_type="contains", negate_condition=True
        )

        # vcard1: has "Alice" - negated, should not match
        self.assertFalse(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: doesn't have "Alice" - negated, should match
        self.assertTrue(filter_obj.check("vcard2.vcf", self.vcard2))

    def test_combined_property_and_text_filters(self):
        """Test combination of property existence and text filtering."""
        filter_obj = CardDAVFilter()
        filter_obj.test = all  # AND logic
        filter_obj.add_property_filter("EMAIL")
        filter_obj.add_property_filter("TEL")
        filter_obj.add_property_filter("ORG").add_text_match(
            "Corp", match_type="contains"
        )

        # vcard1: has EMAIL, TEL, and "Example Corp"
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        # vcard2: no EMAIL
        self.assertFalse(filter_obj.check("vcard2.vcf", self.vcard2))
        # vcard3: has EMAIL and TEL but "Tech Solutions" not "Corp"
        self.assertFalse(filter_obj.check("vcard3.vcf", self.vcard3))

    def test_empty_filter(self):
        """Test filter with no property filters matches all."""
        filter_obj = CardDAVFilter()

        # Should match all vcards
        self.assertTrue(filter_obj.check("vcard1.vcf", self.vcard1))
        self.assertTrue(filter_obj.check("vcard2.vcf", self.vcard2))
        self.assertTrue(filter_obj.check("vcard3.vcf", self.vcard3))


class CardDAVFilterEdgeCasesTests(unittest.TestCase):
    """Test edge cases for CardDAV filters."""

    def test_filter_with_unicode_in_fn(self):
        """Test filtering vCards with Unicode characters."""
        vcard = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Ren\xc3\xa9 Dupont\r\n",
                b"N:Dupont;Ren\xc3\xa9;;;\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "René", collation="i;unicode-casemap", match_type="contains"
        )
        self.assertTrue(filter_obj.check("vcard.vcf", vcard))

    def test_filter_with_empty_property(self):
        """Test filtering when property exists but is empty."""
        vcard = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Test User\r\n",
                b"N:User;Test;;;\r\n",
                b"NOTE:\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Property is defined (even if empty)
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("NOTE")
        self.assertTrue(filter_obj.check("vcard.vcf", vcard))

    def test_filter_multiple_email_addresses(self):
        """Test filtering when multiple instances of same property exist."""
        vcard = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Multi Email\r\n",
                b"N:Email;Multi;;;\r\n",
                b"EMAIL:work@example.com\r\n",
                b"EMAIL:personal@gmail.com\r\n",
                b"EMAIL:other@company.org\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Should match if ANY email matches
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("EMAIL").add_text_match(
            "gmail", collation="i;unicode-casemap", match_type="contains"
        )
        self.assertTrue(filter_obj.check("vcard.vcf", vcard))

    def test_filter_org_structured_property(self):
        """Test filtering on structured ORG property."""
        vcard = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:Structured Org\r\n",
                b"N:Org;Structured;;;\r\n",
                b"ORG:Example Corporation;Engineering Department\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Should match text in any part of structured property
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("ORG").add_text_match(
            "Engineering", match_type="contains"
        )
        self.assertTrue(filter_obj.check("vcard.vcf", vcard))

    def test_filter_n_property(self):
        """Test filtering on N (structured name) property."""
        vcard = VCardFile(
            [
                b"BEGIN:VCARD\r\n",
                b"VERSION:3.0\r\n",
                b"FN:John Doe\r\n",
                b"N:Doe;John;Michael;Dr.;Jr.\r\n",
                b"END:VCARD\r\n",
            ],
            "text/vcard",
        )

        # Should match last name
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("N").add_text_match("Doe", match_type="contains")
        self.assertTrue(filter_obj.check("vcard.vcf", vcard))


if __name__ == "__main__":
    unittest.main()



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Timezone handling.

See http://www.webdav.org/specs/rfc7809.html
"""

from xandikos import webdav


class TimezoneServiceSetProperty(webdav.Property):
    """timezone-service-set property.

    See http://www.webdav.org/specs/rfc7809.html, section 5.1
    """

    name = "{DAV:}timezone-service-set"
    # Should be set on CalDAV calendar home collection resources,
    # but Xandikos doesn't have a separate resource type for those.
    resource_type = webdav.COLLECTION_RESOURCE_TYPE
    in_allprops = False
    live = True

    def __init__(self, timezone_services) -> None:
        super().__init__()
        self._timezone_services = timezone_services

    async def get_value(self, base_href, resource, el, environ):
        for timezone_service_href in self._timezone_services:
            el.append(webdav.create_href(timezone_service_href, base_href))



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.__main__."""

import asyncio
import logging
import os
import shutil
import tempfile
import unittest
from unittest.mock import patch

from xandikos.__main__ import add_create_collection_parser, create_collection_main, main
from xandikos.store import STORE_TYPE_ADDRESSBOOK, STORE_TYPE_CALENDAR
from xandikos.web import SingleUserFilesystemBackend


class CreateCollectionTests(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.test_dir)

    def test_add_create_collection_parser(self):
        """Test that the create-collection parser is correctly configured."""
        import argparse

        parser = argparse.ArgumentParser()
        add_create_collection_parser(parser)

        # Test that required arguments are present
        import sys
        from io import StringIO

        old_stderr = sys.stderr
        sys.stderr = StringIO()  # Suppress argparse error output
        try:
            with self.assertRaises(SystemExit):
                parser.parse_args([])
        finally:
            sys.stderr = old_stderr

        # Test valid arguments
        args = parser.parse_args(
            ["-d", "/test/dir", "--type", "calendar", "--name", "test-cal"]
        )
        self.assertEqual(args.directory, "/test/dir")
        self.assertEqual(args.type, "calendar")
        self.assertEqual(args.name, "test-cal")
        self.assertIsNone(args.displayname)
        self.assertIsNone(args.description)
        self.assertIsNone(args.color)

        # Test with optional arguments
        args = parser.parse_args(
            [
                "-d",
                "/test/dir",
                "--type",
                "addressbook",
                "--name",
                "test-addr",
                "--displayname",
                "Test Address Book",
                "--description",
                "A test address book",
                "--color",
                "#FF0000",
            ]
        )
        self.assertEqual(args.displayname, "Test Address Book")
        self.assertEqual(args.description, "A test address book")
        self.assertEqual(args.color, "#FF0000")

    def test_create_collection_calendar_success(self):
        """Test successful creation of a calendar collection."""
        import argparse

        args = argparse.Namespace(
            directory=self.test_dir,
            type="calendar",
            name="test-calendar",
            displayname="Test Calendar",
            description="A test calendar",
            color="#FF5733",
        )

        result = asyncio.run(create_collection_main(args, None))
        self.assertEqual(result, 0)

        # Verify the collection was created
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "test-calendar")))

        # Verify the collection properties
        backend = SingleUserFilesystemBackend(self.test_dir)
        resource = backend.get_resource("/test-calendar")
        self.assertEqual(resource.store.get_type(), STORE_TYPE_CALENDAR)
        self.assertEqual(resource.store.get_displayname(), "Test Calendar")
        self.assertEqual(resource.store.get_description(), "A test calendar")
        self.assertEqual(resource.store.get_color(), "#FF5733")

    def test_create_collection_addressbook_success(self):
        """Test successful creation of an addressbook collection."""
        import argparse

        args = argparse.Namespace(
            directory=self.test_dir,
            type="addressbook",
            name="test-addressbook",
            displayname="Test Address Book",
            description=None,
            color=None,
        )

        result = asyncio.run(create_collection_main(args, None))
        self.assertEqual(result, 0)

        # Verify the collection was created
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "test-addressbook")))

        # Verify the collection properties
        backend = SingleUserFilesystemBackend(self.test_dir)
        resource = backend.get_resource("/test-addressbook")
        self.assertEqual(resource.store.get_type(), STORE_TYPE_ADDRESSBOOK)
        self.assertEqual(resource.store.get_displayname(), "Test Address Book")

    def test_create_collection_already_exists(self):
        """Test error handling when collection already exists."""
        import argparse

        args = argparse.Namespace(
            directory=self.test_dir,
            type="calendar",
            name="test-calendar",
            displayname=None,
            description=None,
            color=None,
        )

        # Create the collection first time
        result = asyncio.run(create_collection_main(args, None))
        self.assertEqual(result, 0)

        # Try to create again - should fail
        with self.assertLogs("xandikos.__main__", level=logging.ERROR) as cm:
            result = asyncio.run(create_collection_main(args, None))
            self.assertEqual(result, 1)
            self.assertIn("already exists", cm.output[0])

    def test_create_collection_minimal_args(self):
        """Test creating collection with only required arguments."""
        import argparse

        args = argparse.Namespace(
            directory=self.test_dir,
            type="calendar",
            name="minimal-cal",
            displayname=None,
            description=None,
            color=None,
        )

        result = asyncio.run(create_collection_main(args, None))
        self.assertEqual(result, 0)

        # Verify the collection was created
        backend = SingleUserFilesystemBackend(self.test_dir)
        resource = backend.get_resource("/minimal-cal")
        self.assertEqual(resource.store.get_type(), STORE_TYPE_CALENDAR)


class MainCommandTests(unittest.TestCase):
    def test_main_create_collection_subcommand(self):
        """Test that the main function recognizes create-collection subcommand."""
        with tempfile.TemporaryDirectory() as test_dir:
            with self.assertLogs("xandikos.__main__", level=logging.INFO) as cm:
                result = asyncio.run(
                    main(
                        [
                            "create-collection",
                            "-d",
                            test_dir,
                            "--type",
                            "calendar",
                            "--name",
                            "test-cal",
                        ]
                    )
                )
                self.assertEqual(result, 0)
                self.assertIn("Successfully created", cm.output[0])

    def test_main_help_includes_create_collection(self):
        """Test that help includes the create-collection subcommand."""
        import sys
        from io import StringIO

        # Capture stdout since argparse writes help there
        old_stdout = sys.stdout
        captured_output = StringIO()
        sys.stdout = captured_output

        try:
            with self.assertRaises(SystemExit):
                asyncio.run(main(["--help"]))
        finally:
            sys.stdout = old_stdout

        # Check that help was printed and includes create-collection
        help_output = captured_output.getvalue()
        self.assertIn("create-collection", help_output)

    def test_main_invalid_subcommand(self):
        """Test handling of invalid subcommands."""
        # Note: Due to the default subparser mechanism, unknown commands
        # get passed to the 'serve' subcommand and cause argument errors.
        # This test verifies that the system exits with an error code.
        import sys
        from io import StringIO

        old_stderr = sys.stderr
        sys.stderr = StringIO()  # Suppress argparse error output
        try:
            with patch("builtins.print"):
                with self.assertRaises(SystemExit) as cm:
                    asyncio.run(main(["invalid-command"]))
                # Expect exit code 2 (argparse error)
                self.assertEqual(cm.exception.code, 2)
        finally:
            sys.stderr = old_stderr

    def test_main_create_collection_help(self):
        """Test create-collection subcommand help."""
        import sys
        from io import StringIO

        # Capture stdout since argparse writes help there
        old_stdout = sys.stdout
        captured_output = StringIO()
        sys.stdout = captured_output

        try:
            with self.assertRaises(SystemExit):
                asyncio.run(main(["create-collection", "--help"]))
        finally:
            sys.stdout = old_stdout

        # Check that help was printed and includes expected options
        help_output = captured_output.getvalue()
        self.assertIn("--type", help_output)
        self.assertIn("--name", help_output)
        self.assertIn("--displayname", help_output)

    def test_main_help_subcommand(self):
        """Test that 'help' subcommand prints usage and returns 0."""
        import sys
        from io import StringIO

        old_stdout = sys.stdout
        captured_output = StringIO()
        sys.stdout = captured_output

        try:
            result = asyncio.run(main(["help"]))
        finally:
            sys.stdout = old_stdout

        self.assertEqual(result, 0)
        help_output = captured_output.getvalue()
        # Should list all subcommands
        self.assertIn("serve", help_output)
        self.assertIn("create-collection", help_output)
        self.assertIn("help", help_output)



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.multi_user module."""

import argparse
import os
import shutil
import tempfile
import unittest

from xandikos.multi_user import (
    MultiUserFilesystemBackend,
    MultiUserXandikosApp,
    InvalidUsernameError,
    validate_username,
    add_parser,
)
from xandikos.store import (
    STORE_TYPE_ADDRESSBOOK,
    STORE_TYPE_CALENDAR,
    STORE_TYPE_SCHEDULE_INBOX,
)
from xandikos.webdav import ForbiddenError


class MultiUserFilesystemBackendTests(unittest.TestCase):
    """Tests for MultiUserFilesystemBackend."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_init_default_paths(self):
        """Test initialization with default path settings."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        self.assertEqual(backend.principal_path_prefix, "/")
        self.assertEqual(backend.principal_path_suffix, "/")
        self.assertTrue(backend.autocreate)

    def test_init_custom_paths(self):
        """Test initialization with custom path settings."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            principal_path_prefix="/users/",
            principal_path_suffix="/dav/",
        )

        self.assertEqual(backend.principal_path_prefix, "/users/")
        self.assertEqual(backend.principal_path_suffix, "/dav/")

    def test_init_with_kwargs(self):
        """Test that additional kwargs are passed to parent class."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            paranoid=True,
            index_threshold=100,
        )

        self.assertTrue(backend.paranoid)
        self.assertEqual(backend.index_threshold, 100)

    def test_show_principals_on_root_default(self):
        """Test that show_principals_on_root defaults to True."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        self.assertTrue(backend.show_principals_on_root)

    def test_show_principals_on_root_can_be_disabled(self):
        """Test that show_principals_on_root can be set to False."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            show_principals_on_root=False,
        )
        self.assertFalse(backend.show_principals_on_root)

    def test_set_principal_creates_principal(self):
        """Test that set_principal creates a new principal."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("alice")

        resource = backend.get_resource("/alice/")
        self.assertIsNotNone(resource)

    def test_set_principal_marks_as_principal(self):
        """Test that set_principal marks the path as a principal."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("alice")

        # _mark_as_principal normalizes paths
        self.assertIn("/alice", backend._user_principals)

    def test_set_principal_with_custom_prefix(self):
        """Test set_principal with custom prefix."""
        backend = MultiUserFilesystemBackend(
            self.test_dir, principal_path_prefix="/users/"
        )
        backend.set_principal("bob")

        resource = backend.get_resource("/users/bob/")
        self.assertIsNotNone(resource)
        self.assertIn("/users/bob", backend._user_principals)

    def test_set_principal_with_custom_suffix(self):
        """Test set_principal with custom suffix."""
        backend = MultiUserFilesystemBackend(
            self.test_dir, principal_path_suffix="/principal/"
        )
        backend.set_principal("charlie")

        resource = backend.get_resource("/charlie/principal/")
        self.assertIsNotNone(resource)
        self.assertIn("/charlie/principal", backend._user_principals)

    def test_set_principal_with_custom_prefix_and_suffix(self):
        """Test set_principal with both custom prefix and suffix."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            principal_path_prefix="/accounts/",
            principal_path_suffix="/home/",
        )
        backend.set_principal("dave")

        resource = backend.get_resource("/accounts/dave/home/")
        self.assertIsNotNone(resource)
        self.assertIn("/accounts/dave/home", backend._user_principals)

    def test_set_principal_override_prefix(self):
        """Test overriding prefix at method call time."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("eve", principal_path_prefix="/special/")

        resource = backend.get_resource("/special/eve/")
        self.assertIsNotNone(resource)
        self.assertIn("/special/eve", backend._user_principals)

    def test_set_principal_override_suffix(self):
        """Test overriding suffix at method call time."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("frank", principal_path_suffix="/user/")

        resource = backend.get_resource("/frank/user/")
        self.assertIsNotNone(resource)
        self.assertIn("/frank/user", backend._user_principals)

    def test_set_principal_override_both(self):
        """Test overriding both prefix and suffix at method call time."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal(
            "grace",
            principal_path_prefix="/custom/",
            principal_path_suffix="/path/",
        )

        resource = backend.get_resource("/custom/grace/path/")
        self.assertIsNotNone(resource)
        self.assertIn("/custom/grace/path", backend._user_principals)

    def test_set_principal_idempotent(self):
        """Test that calling set_principal twice is idempotent."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("henry")
        resource1 = backend.get_resource("/henry/")

        backend.set_principal("henry")
        resource2 = backend.get_resource("/henry/")

        self.assertIsNotNone(resource1)
        self.assertIsNotNone(resource2)

    def test_set_principal_multiple_users(self):
        """Test creating multiple users."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        users = ["alice", "bob", "charlie", "dave", "eve"]
        for user in users:
            backend.set_principal(user)

        for user in users:
            resource = backend.get_resource(f"/{user}/")
            self.assertIsNotNone(resource)
            self.assertIn(f"/{user}", backend._user_principals)

    def test_set_principal_creates_defaults(self):
        """Test that set_principal creates default calendar and addressbook."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("ivan")

        # Check calendar was created
        calendar = backend.get_resource("/ivan/calendars/calendar/")
        self.assertIsNotNone(calendar)
        self.assertEqual(calendar.store.get_type(), STORE_TYPE_CALENDAR)

        # Check addressbook was created
        addressbook = backend.get_resource("/ivan/contacts/addressbook/")
        self.assertIsNotNone(addressbook)
        self.assertEqual(addressbook.store.get_type(), STORE_TYPE_ADDRESSBOOK)

        # Check inbox was created
        inbox = backend.get_resource("/ivan/inbox/")
        self.assertIsNotNone(inbox)
        self.assertEqual(inbox.store.get_type(), STORE_TYPE_SCHEDULE_INBOX)

    def test_set_principal_with_special_characters(self):
        """Test usernames with special characters."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        # Test with underscores
        backend.set_principal("john_doe")
        self.assertIsNotNone(backend.get_resource("/john_doe/"))

        # Test with hyphens
        backend.set_principal("jane-doe")
        self.assertIsNotNone(backend.get_resource("/jane-doe/"))

        # Test with dots
        backend.set_principal("user.name")
        self.assertIsNotNone(backend.get_resource("/user.name/"))

    def test_set_principal_case_sensitive(self):
        """Test that usernames are case-sensitive."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("Alice")
        backend.set_principal("alice")

        # Both should exist as separate principals
        self.assertIn("/Alice", backend._user_principals)
        self.assertIn("/alice", backend._user_principals)

    def test_find_principals_empty(self):
        """Test find_principals returns empty set initially."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        self.assertEqual(backend.find_principals(), set())

    def test_find_principals_after_set(self):
        """Test find_principals returns created principals."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("user1")
        backend.set_principal("user2")

        principals = backend.find_principals()
        self.assertIn("/user1", principals)
        self.assertIn("/user2", principals)

    def test_inherits_from_xandikos_backend(self):
        """Test that MultiUserFilesystemBackend has XandikosBackend methods."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        # Should have methods from XandikosBackend
        self.assertTrue(hasattr(backend, "get_resource"))
        self.assertTrue(hasattr(backend, "create_collection"))
        self.assertTrue(hasattr(backend, "create_principal"))
        self.assertTrue(hasattr(backend, "_mark_as_principal"))

    def test_create_collection_after_principal(self):
        """Test creating additional collections after principal setup."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("kate")

        # Create an additional calendar
        collection = backend.create_collection("/kate/calendars/work/")
        self.assertIsNotNone(collection)

    def test_empty_username_rejected(self):
        """Test that empty usernames are rejected for security."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        with self.assertRaises(InvalidUsernameError):
            backend.set_principal("")


class AddParserTests(unittest.TestCase):
    """Tests for add_parser function."""

    def test_add_parser_creates_arguments(self):
        """Test that add_parser adds expected arguments."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        # Parse with required arguments
        args = parser.parse_args(["-d", "/tmp/test"])

        self.assertEqual(args.directory, "/tmp/test")
        self.assertEqual(args.principal_path_prefix, "/")
        self.assertEqual(args.principal_path_suffix, "/")
        self.assertEqual(args.listen_address, "localhost")
        self.assertEqual(args.port, 8080)
        self.assertFalse(args.dump_dav_xml)
        self.assertFalse(args.avahi)
        self.assertTrue(args.strict)
        self.assertFalse(args.debug)

    def test_add_parser_directory_required(self):
        """Test that directory argument is required."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        with self.assertRaises(SystemExit):
            parser.parse_args([])

    def test_add_parser_custom_principal_paths(self):
        """Test parsing custom principal path arguments."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "--principal-path-prefix",
                "/users/",
                "--principal-path-suffix",
                "/dav/",
            ]
        )

        self.assertEqual(args.principal_path_prefix, "/users/")
        self.assertEqual(args.principal_path_suffix, "/dav/")

    def test_add_parser_listen_options(self):
        """Test parsing listen address and port."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "-l",
                "0.0.0.0",
                "-p",
                "9090",
            ]
        )

        self.assertEqual(args.listen_address, "0.0.0.0")
        self.assertEqual(args.port, 9090)

    def test_add_parser_long_listen_options(self):
        """Test parsing long-form listen options."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "--directory",
                "/tmp/test",
                "--listen-address",
                "127.0.0.1",
                "--port",
                "8888",
            ]
        )

        self.assertEqual(args.directory, "/tmp/test")
        self.assertEqual(args.listen_address, "127.0.0.1")
        self.assertEqual(args.port, 8888)

    def test_add_parser_socket_options(self):
        """Test parsing unix socket options."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "--socket-mode",
                "660",
                "--socket-group",
                "www-data",
            ]
        )

        self.assertEqual(args.socket_mode, "660")
        self.assertEqual(args.socket_group, "www-data")

    def test_add_parser_route_prefix(self):
        """Test parsing route prefix."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "--route-prefix",
                "/caldav/",
            ]
        )

        self.assertEqual(args.route_prefix, "/caldav/")

    def test_add_parser_metrics_port(self):
        """Test parsing metrics port."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "--metrics-port",
                "9100",
            ]
        )

        self.assertEqual(args.metrics_port, "9100")

    def test_add_parser_debug_flag(self):
        """Test parsing debug flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--debug"])
        self.assertTrue(args.debug)

    def test_add_parser_dump_dav_xml_flag(self):
        """Test parsing dump-dav-xml flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--dump-dav-xml"])
        self.assertTrue(args.dump_dav_xml)

    def test_add_parser_avahi_flag(self):
        """Test parsing avahi flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--avahi"])
        self.assertTrue(args.avahi)

    def test_add_parser_no_strict_flag(self):
        """Test parsing no-strict flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--no-strict"])
        self.assertFalse(args.strict)

    def test_add_parser_no_detect_systemd_flag(self):
        """Test parsing no-detect-systemd flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--no-detect-systemd"])
        self.assertFalse(args.detect_systemd)

    def test_add_parser_hidden_paranoid(self):
        """Test parsing hidden paranoid argument."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--paranoid"])
        self.assertTrue(args.paranoid)

    def test_add_parser_hidden_index_threshold(self):
        """Test parsing hidden index-threshold argument."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(["-d", "/tmp/test", "--index-threshold", "500"])
        self.assertEqual(args.index_threshold, 500)

    def test_add_parser_unix_socket_path(self):
        """Test parsing unix socket as listen address."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        args = parser.parse_args(
            [
                "-d",
                "/tmp/test",
                "-l",
                "/var/run/xandikos.sock",
            ]
        )

        self.assertEqual(args.listen_address, "/var/run/xandikos.sock")

    def test_add_parser_hide_principals_flag(self):
        """Test parsing hide-principals flag."""
        parser = argparse.ArgumentParser()
        add_parser(parser)

        # Default is False (show principals)
        args = parser.parse_args(["-d", "/tmp/test"])
        self.assertFalse(args.hide_principals)

        # With flag, hide principals
        args = parser.parse_args(["-d", "/tmp/test", "--hide-principals"])
        self.assertTrue(args.hide_principals)


class MultiUserIntegrationTests(unittest.TestCase):
    """Integration tests for multi-user functionality."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_user_isolation(self):
        """Test that users have isolated data spaces."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("user1")
        backend.set_principal("user2")

        # Each user should have their own calendar home
        user1_calendars = backend.get_resource("/user1/calendars/")
        user2_calendars = backend.get_resource("/user2/calendars/")

        self.assertIsNotNone(user1_calendars)
        self.assertIsNotNone(user2_calendars)

        # Verify they are different resources
        self.assertNotEqual(
            backend._map_to_file_path("/user1/calendars/"),
            backend._map_to_file_path("/user2/calendars/"),
        )

    def test_user_directory_structure(self):
        """Test that correct directory structure is created."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("testuser")

        # Check directory structure exists on disk
        user_path = os.path.join(self.test_dir, "testuser")
        self.assertTrue(os.path.isdir(user_path))

        calendars_path = os.path.join(user_path, "calendars")
        self.assertTrue(os.path.isdir(calendars_path))

        contacts_path = os.path.join(user_path, "contacts")
        self.assertTrue(os.path.isdir(contacts_path))

    def test_nested_principal_paths(self):
        """Test principals with deeply nested paths."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            principal_path_prefix="/org/company/users/",
            principal_path_suffix="/profile/",
        )
        backend.set_principal("employee")

        resource = backend.get_resource("/org/company/users/employee/profile/")
        self.assertIsNotNone(resource)

        # Check file system path
        expected_path = os.path.join(
            self.test_dir, "org", "company", "users", "employee", "profile"
        )
        self.assertTrue(os.path.isdir(expected_path))

    def test_concurrent_user_creation(self):
        """Test creating many users."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        # Create many users
        for i in range(50):
            backend.set_principal(f"user{i}")

        # Verify all were created
        for i in range(50):
            resource = backend.get_resource(f"/user{i}/")
            self.assertIsNotNone(resource)

        self.assertEqual(len(backend.find_principals()), 50)

    def test_principal_with_numeric_name(self):
        """Test principal with numeric username."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("12345")

        resource = backend.get_resource("/12345/")
        self.assertIsNotNone(resource)

    def test_principal_calendar_operations(self):
        """Test basic calendar operations after principal creation."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("caluser")

        calendar = backend.get_resource("/caluser/calendars/calendar/")
        self.assertIsNotNone(calendar)

        # Test setting calendar properties
        calendar.store.set_description("Test Calendar")
        self.assertEqual(calendar.store.get_description(), "Test Calendar")

        calendar.store.set_color("#FF0000")
        self.assertEqual(calendar.store.get_color(), "#FF0000")

    def test_principal_addressbook_operations(self):
        """Test basic addressbook operations after principal creation."""
        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("addruser")

        addressbook = backend.get_resource("/addruser/contacts/addressbook/")
        self.assertIsNotNone(addressbook)

        # Test setting addressbook properties
        addressbook.store.set_description("Test Addressbook")
        self.assertEqual(addressbook.store.get_description(), "Test Addressbook")


class UserAccessControlTests(unittest.TestCase):
    """Tests for user access control and isolation."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_current_user_principal_substitution(self):
        """Test that current_user_principal uses REMOTE_USER substitution."""
        from xandikos.web import XandikosApp

        backend = MultiUserFilesystemBackend(self.test_dir)
        backend.set_principal("alice")

        # Create app with template for current_user_principal
        app = XandikosApp(
            backend,
            current_user_principal="/%(REMOTE_USER)s/",
        )

        # The app should be configured with the template
        self.assertIsNotNone(app)

    def test_users_have_separate_calendar_homes(self):
        """Test that each user has a separate calendar home path."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")

        alice_calendar = backend.get_resource("/alice/calendars/calendar/")
        bob_calendar = backend.get_resource("/bob/calendars/calendar/")

        self.assertIsNotNone(alice_calendar)
        self.assertIsNotNone(bob_calendar)

        # Verify they have different file paths
        self.assertNotEqual(
            alice_calendar.store.path,
            bob_calendar.store.path,
        )

    def test_users_have_separate_addressbook_homes(self):
        """Test that each user has a separate addressbook home path."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")

        alice_addressbook = backend.get_resource("/alice/contacts/addressbook/")
        bob_addressbook = backend.get_resource("/bob/contacts/addressbook/")

        self.assertIsNotNone(alice_addressbook)
        self.assertIsNotNone(bob_addressbook)

        # Verify they have different file paths
        self.assertNotEqual(
            alice_addressbook.store.path,
            bob_addressbook.store.path,
        )

    def test_user_data_stored_in_separate_directories(self):
        """Test that user data is stored in physically separate directories."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")

        alice_path = os.path.join(self.test_dir, "alice")
        bob_path = os.path.join(self.test_dir, "bob")

        self.assertTrue(os.path.isdir(alice_path))
        self.assertTrue(os.path.isdir(bob_path))

        # Verify no overlap in paths
        self.assertFalse(alice_path.startswith(bob_path))
        self.assertFalse(bob_path.startswith(alice_path))

    def test_backend_can_access_any_user_resource(self):
        """Test that backend access is not restricted by user.

        NOTE: This documents current behavior - the backend does not
        enforce access control. This is intentional as access control
        is expected to be handled by a reverse proxy.
        """
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")

        # Backend can access both users' resources
        alice_calendar = backend.get_resource("/alice/calendars/calendar/")
        bob_calendar = backend.get_resource("/bob/calendars/calendar/")

        self.assertIsNotNone(alice_calendar)
        self.assertIsNotNone(bob_calendar)

    def test_principal_path_determines_user_home(self):
        """Test that principal path correctly determines user's home location."""
        backend = MultiUserFilesystemBackend(
            self.test_dir,
            principal_path_prefix="/users/",
            principal_path_suffix="/",
        )

        backend.set_principal("alice")

        # Alice's resources should be under /users/alice/
        alice_principal = backend.get_resource("/users/alice/")
        alice_calendars = backend.get_resource("/users/alice/calendars/")

        self.assertIsNotNone(alice_principal)
        self.assertIsNotNone(alice_calendars)

        # Root should not contain alice directly
        root = backend.get_resource("/")
        self.assertIsNotNone(root)

    def test_find_principals_returns_all_users(self):
        """Test that find_principals returns all registered users."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")
        backend.set_principal("charlie")

        principals = backend.find_principals()

        self.assertEqual(len(principals), 3)
        self.assertIn("/alice", principals)
        self.assertIn("/bob", principals)
        self.assertIn("/charlie", principals)

    def test_each_user_gets_own_inbox(self):
        """Test that each user has their own schedule inbox."""
        backend = MultiUserFilesystemBackend(self.test_dir)

        backend.set_principal("alice")
        backend.set_principal("bob")

        alice_inbox = backend.get_resource("/alice/inbox/")
        bob_inbox = backend.get_resource("/bob/inbox/")

        self.assertIsNotNone(alice_inbox)
        self.assertIsNotNone(bob_inbox)

        # Verify they have different file paths
        self.assertNotEqual(
            alice_inbox.store.path,
            bob_inbox.store.path,
        )


class MultiUserXandikosAppAuthorizationTests(unittest.TestCase):
    """Tests for MultiUserXandikosApp path-based authorization."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.backend = MultiUserFilesystemBackend(self.test_dir)
        self.backend.set_principal("alice")
        self.backend.set_principal("bob")
        self.app = MultiUserXandikosApp(
            self.backend,
            current_user_principal="/%(REMOTE_USER)s/",
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_user_can_access_own_principal(self):
        """Test that a user can access their own principal path."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/alice/", "GET")
        self.app.check_access(environ, "/alice", "GET")

    def test_user_can_access_own_calendar(self):
        """Test that a user can access their own calendar."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/alice/calendars/", "GET")
        self.app.check_access(environ, "/alice/calendars/calendar/", "GET")

    def test_user_can_access_own_addressbook(self):
        """Test that a user can access their own addressbook."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/alice/contacts/", "GET")
        self.app.check_access(environ, "/alice/contacts/addressbook/", "GET")

    def test_user_cannot_access_other_user_principal(self):
        """Test that a user cannot access another user's principal."""
        environ = {"REMOTE_USER": "alice"}
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/", "GET")

    def test_user_cannot_access_other_user_calendar(self):
        """Test that a user cannot access another user's calendar."""
        environ = {"REMOTE_USER": "alice"}
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/", "GET")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/calendar/", "GET")

    def test_user_cannot_access_other_user_addressbook(self):
        """Test that a user cannot access another user's addressbook."""
        environ = {"REMOTE_USER": "alice"}
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/contacts/", "GET")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/contacts/addressbook/", "GET")

    def test_user_can_access_root(self):
        """Test that users can access the root path for discovery."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/", "GET")

    def test_user_can_access_wellknown_caldav(self):
        """Test that users can access .well-known/caldav."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/.well-known/caldav", "GET")

    def test_user_can_access_wellknown_carddav(self):
        """Test that users can access .well-known/carddav."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/.well-known/carddav", "GET")

    def test_unauthenticated_user_denied_by_default(self):
        """Test that unauthenticated requests are denied by default.

        Multi-user mode requires authentication to access user resources.
        Only root and well-known paths are accessible without authentication.
        """
        environ = {}  # No REMOTE_USER
        # Should raise ForbiddenError for user paths
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/alice/calendars/", "GET")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/", "GET")

    def test_authorization_applies_to_all_methods(self):
        """Test that authorization applies to various HTTP methods."""
        environ = {"REMOTE_USER": "alice"}

        # Alice can use various methods on her own resources
        for method in ["GET", "PUT", "DELETE", "PROPFIND", "PROPPATCH", "MKCOL"]:
            self.app.check_access(environ, "/alice/calendars/", method)

        # Alice cannot use any method on Bob's resources
        for method in ["GET", "PUT", "DELETE", "PROPFIND", "PROPPATCH", "MKCOL"]:
            with self.assertRaises(ForbiddenError):
                self.app.check_access(environ, "/bob/calendars/", method)

    def test_forbidden_error_message_is_generic(self):
        """Test that ForbiddenError uses generic message to avoid info disclosure."""
        environ = {"REMOTE_USER": "alice"}
        try:
            self.app.check_access(environ, "/bob/calendars/", "GET")
            self.fail("Expected ForbiddenError")
        except ForbiddenError as e:
            # Should be generic message without username or path details
            self.assertEqual(e.message, "Access denied")
            # Should NOT contain sensitive details
            self.assertNotIn("alice", e.message)
            self.assertNotIn("/bob/calendars/", e.message)

    def test_path_normalization(self):
        """Test that paths are normalized correctly."""
        environ = {"REMOTE_USER": "alice"}

        # Various path formats should work for Alice
        self.app.check_access(environ, "/alice", "GET")
        self.app.check_access(environ, "/alice/", "GET")
        self.app.check_access(environ, "alice/", "GET")  # Without leading slash

        # Various path formats should be denied for Bob
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob", "GET")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/", "GET")


class MultiUserXandikosAppCustomPrefixTests(unittest.TestCase):
    """Tests for MultiUserXandikosApp with custom principal path prefix/suffix."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.backend = MultiUserFilesystemBackend(
            self.test_dir,
            principal_path_prefix="/users/",
            principal_path_suffix="/home/",
        )
        self.backend.set_principal("alice")
        self.backend.set_principal("bob")
        self.app = MultiUserXandikosApp(
            self.backend,
            current_user_principal="/users/%(REMOTE_USER)s/home/",
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_user_can_access_own_path_with_custom_prefix(self):
        """Test access with custom principal path prefix."""
        environ = {"REMOTE_USER": "alice"}
        # Should not raise
        self.app.check_access(environ, "/users/alice/home/", "GET")
        self.app.check_access(environ, "/users/alice/home/calendars/", "GET")

    def test_user_cannot_access_other_path_with_custom_prefix(self):
        """Test denial with custom principal path prefix."""
        environ = {"REMOTE_USER": "alice"}
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/users/bob/home/", "GET")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/users/bob/home/calendars/", "GET")

    def test_user_cannot_access_partial_match(self):
        """Test that partial path matches don't grant access."""
        environ = {"REMOTE_USER": "alice"}
        # /users/alice/ without /home/ suffix should be denied
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/users/alice/", "GET")
        # /alice/ without /users/ prefix should be denied
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/alice/home/", "GET")

    def test_root_still_accessible(self):
        """Test that root is still accessible with custom prefix."""
        environ = {"REMOTE_USER": "alice"}
        self.app.check_access(environ, "/", "GET")


class UsernameValidationTests(unittest.TestCase):
    """Tests for username validation security."""

    def test_valid_username(self):
        """Test that valid usernames are accepted."""
        # Should not raise
        validate_username("alice")
        validate_username("bob123")
        validate_username("user-name")
        validate_username("user_name")
        validate_username("user.name")
        validate_username("UPPERCASE")

    def test_empty_username_rejected(self):
        """Test that empty usernames are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("")

    def test_slash_in_username_rejected(self):
        """Test that usernames with slashes are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("user/name")
        with self.assertRaises(InvalidUsernameError):
            validate_username("/alice")
        with self.assertRaises(InvalidUsernameError):
            validate_username("alice/")

    def test_backslash_in_username_rejected(self):
        """Test that usernames with backslashes are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("user\\name")

    def test_dotdot_in_username_rejected(self):
        """Test that usernames with '..' are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("..")
        with self.assertRaises(InvalidUsernameError):
            validate_username("..bob")
        with self.assertRaises(InvalidUsernameError):
            validate_username("bob..")
        with self.assertRaises(InvalidUsernameError):
            validate_username("a..b")

    def test_null_byte_in_username_rejected(self):
        """Test that usernames with null bytes are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("user\x00name")

    def test_long_username_rejected(self):
        """Test that extremely long usernames are rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username("a" * 256)

    def test_dot_username_rejected(self):
        """Test that '.' username is rejected."""
        with self.assertRaises(InvalidUsernameError):
            validate_username(".")

    def test_max_length_username_accepted(self):
        """Test that max length usernames are accepted."""
        # Should not raise
        validate_username("a" * 255)


class SecurityTests(unittest.TestCase):
    """Security-focused tests for multi-user mode."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.backend = MultiUserFilesystemBackend(self.test_dir)
        self.backend.set_principal("alice")
        self.backend.set_principal("bob")
        self.app = MultiUserXandikosApp(
            self.backend,
            current_user_principal="/%(REMOTE_USER)s/",
            require_auth=True,
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_unauthenticated_access_to_user_data_denied(self):
        """Test that unauthenticated requests cannot access user data."""
        environ = {}  # No REMOTE_USER
        with self.assertRaises(ForbiddenError) as ctx:
            self.app.check_access(environ, "/alice/calendars/", "GET")
        # Message should be generic to avoid path disclosure
        self.assertEqual(ctx.exception.message, "Authentication required")

    def test_unauthenticated_can_access_root(self):
        """Test that unauthenticated requests can access root for discovery."""
        environ = {}  # No REMOTE_USER
        # Should not raise
        self.app.check_access(environ, "/", "GET")

    def test_unauthenticated_can_access_wellknown(self):
        """Test that unauthenticated requests can access well-known paths."""
        environ = {}  # No REMOTE_USER
        # Should not raise
        self.app.check_access(environ, "/.well-known/caldav", "GET")
        self.app.check_access(environ, "/.well-known/carddav", "GET")

    def test_path_traversal_username_rejected(self):
        """Test that path traversal attempts in username are rejected."""
        with self.assertRaises(InvalidUsernameError):
            self.backend.set_principal("../bob")
        with self.assertRaises(InvalidUsernameError):
            self.backend.set_principal("alice/../bob")
        with self.assertRaises(InvalidUsernameError):
            self.backend.set_principal("alice/../../etc/passwd")

    def test_require_auth_false_allows_unauthenticated(self):
        """Test that require_auth=False allows unauthenticated access."""
        app = MultiUserXandikosApp(
            self.backend,
            current_user_principal="/%(REMOTE_USER)s/",
            require_auth=False,
        )
        environ = {}  # No REMOTE_USER
        # Should not raise with require_auth=False
        app.check_access(environ, "/alice/calendars/", "GET")

    def test_cannot_access_other_user_via_move_destination(self):
        """Test that MOVE destination to other user's path is denied.

        This tests that the authorization is checked on destination paths,
        not just source paths.
        """
        environ = {"REMOTE_USER": "alice"}
        # Alice can access her own path
        self.app.check_access(environ, "/alice/calendars/event.ics", "MOVE")
        # But not Bob's path (even as destination)
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/event.ics", "MOVE")

    def test_cannot_access_other_user_via_copy_destination(self):
        """Test that COPY destination to other user's path is denied."""
        environ = {"REMOTE_USER": "alice"}
        # Alice can access her own path
        self.app.check_access(environ, "/alice/calendars/event.ics", "COPY")
        # But not Bob's path (even as destination)
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/event.ics", "COPY")

    def test_all_http_methods_authorized(self):
        """Test that authorization applies to all HTTP methods."""
        environ = {"REMOTE_USER": "alice"}
        methods = [
            "GET",
            "PUT",
            "DELETE",
            "POST",
            "PROPFIND",
            "PROPPATCH",
            "MKCOL",
            "MKCALENDAR",
            "MOVE",
            "COPY",
            "REPORT",
            "OPTIONS",
        ]
        for method in methods:
            # Alice can use any method on her resources
            self.app.check_access(environ, "/alice/calendars/", method)
            # But not on Bob's
            with self.assertRaises(ForbiddenError):
                self.app.check_access(environ, "/bob/calendars/", method)


class PropfindAuthorizationTests(unittest.TestCase):
    """Tests for PROPFIND authorization filtering with depth > 0."""

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.backend = MultiUserFilesystemBackend(self.test_dir)
        self.backend.set_principal("alice")
        self.backend.set_principal("bob")
        self.app = MultiUserXandikosApp(
            self.backend,
            current_user_principal="/%(REMOTE_USER)s/",
            require_auth=True,
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_propfind_on_root_only_shows_own_principal(self):
        """Test that PROPFIND on root with depth only shows user's own principal.

        This is a critical security test - PROPFIND with Depth > 0 on root
        should only return resources the user is authorized to see, not
        other users' resources.
        """
        # Create environ for alice
        environ = {"REMOTE_USER": "alice", "SCRIPT_NAME": ""}

        # Test the check_access callback mechanism used by traverse_resource
        # Alice should be able to access her own paths
        self.app.check_access(environ, "/alice/", "PROPFIND")
        self.app.check_access(environ, "/alice/calendars/", "PROPFIND")

        # Alice should NOT be able to access Bob's paths
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/", "PROPFIND")
        with self.assertRaises(ForbiddenError):
            self.app.check_access(environ, "/bob/calendars/", "PROPFIND")

    def test_propfind_depth_filtering_callback(self):
        """Test the filtering callback pattern used by PropfindMethod."""
        from xandikos.webdav import ForbiddenError, href_to_path

        environ = {"REMOTE_USER": "alice", "SCRIPT_NAME": ""}

        def check_resource_access(href: str) -> bool:
            """Simulate the check used in PropfindMethod."""
            path = href_to_path(environ, href)
            if path is None:
                return False
            try:
                self.app.check_access(environ, path, "PROPFIND")
                return True
            except ForbiddenError:
                return False

        # Alice's paths should pass
        self.assertTrue(check_resource_access("/alice/"))
        self.assertTrue(check_resource_access("/alice/calendars/"))
        self.assertTrue(check_resource_access("/alice/contacts/"))

        # Bob's paths should fail
        self.assertFalse(check_resource_access("/bob/"))
        self.assertFalse(check_resource_access("/bob/calendars/"))
        self.assertFalse(check_resource_access("/bob/contacts/"))

        # Root should pass (allowed for discovery)
        self.assertTrue(check_resource_access("/"))

    def test_traverse_resource_with_authorization(self):
        """Test traverse_resource with authorization filtering.

        This test verifies that traverse_resource correctly filters out
        resources that the user is not authorized to access, using Alice's
        calendar collection as the base resource.
        """
        import asyncio
        from xandikos.webdav import traverse_resource

        # Get alice's calendar collection (which is a real Collection)
        alice_calendars = self.backend.get_resource("/alice/calendars/")
        self.assertIsNotNone(alice_calendars)

        environ = {"REMOTE_USER": "alice", "SCRIPT_NAME": ""}

        def check_access(href: str) -> bool:
            """Authorization callback for traverse_resource."""
            path = href.rstrip("/") or "/"
            if not path.startswith("/"):
                path = "/" + path
            try:
                self.app.check_access(environ, path, "PROPFIND")
                return True
            except ForbiddenError:
                return False

        async def collect_traversed():
            results = []
            async for href, resource in traverse_resource(
                alice_calendars, "/alice/calendars/", "1", check_access=check_access
            ):
                results.append(href)
            return results

        traversed = asyncio.run(collect_traversed())

        # Should include alice's calendars collection
        self.assertIn("/alice/calendars/", traversed)

        # Should include alice's default calendar
        self.assertIn("/alice/calendars/calendar/", traversed)

    def test_traverse_filters_unauthorized_children(self):
        """Test that traverse_resource skips unauthorized children.

        This simulates the scenario where a collection contains references
        to resources the user cannot access (e.g., if we had a shared
        calendar home with multiple users' calendars).
        """
        import asyncio
        from xandikos.webdav import traverse_resource, Collection

        # Create a mock collection that returns both alice and bob resources
        class MockCollection(Collection):
            resource_types = Collection.resource_types

            def members(self):
                # Simulate a collection that contains both users' resources
                alice_cal = self.backend.get_resource("/alice/calendars/calendar/")
                bob_cal = self.backend.get_resource("/bob/calendars/calendar/")
                # Note: in real code paths would be relative, but we use
                # absolute hrefs here for testing the authorization filter
                yield ("alice-calendar", alice_cal)
                yield ("bob-calendar", bob_cal)

            def __init__(self, backend):
                self.backend = backend

        mock_collection = MockCollection(self.backend)

        def check_access(href: str) -> bool:
            """Check if alice can access the path."""
            # For this test, alice can only access her own resources
            # The href will be like /shared/bob-calendar/ (with trailing slash for collections)
            if "bob-calendar" in href:
                return False
            return True

        async def collect_traversed():
            results = []
            async for href, resource in traverse_resource(
                mock_collection, "/shared/", "1", check_access=check_access
            ):
                results.append(href)
            return results

        traversed = asyncio.run(collect_traversed())

        # Should include the base collection
        self.assertIn("/shared/", traversed)

        # Should include alice's calendar (authorized) - collections get trailing slash
        self.assertIn("/shared/alice-calendar/", traversed)

        # Should NOT include bob's calendar (filtered out)
        self.assertNotIn("/shared/bob-calendar/", traversed)


class ModuleExportsTests(unittest.TestCase):
    """Tests for module exports and imports."""

    def test_module_exports(self):
        """Test that expected symbols are exported."""
        from xandikos import multi_user

        self.assertTrue(hasattr(multi_user, "MultiUserFilesystemBackend"))
        self.assertTrue(hasattr(multi_user, "MultiUserXandikosApp"))
        self.assertTrue(hasattr(multi_user, "InvalidUsernameError"))
        self.assertTrue(hasattr(multi_user, "validate_username"))
        self.assertTrue(hasattr(multi_user, "add_parser"))
        self.assertTrue(hasattr(multi_user, "main"))

    def test_all_exports(self):
        """Test __all__ contains expected exports."""
        from xandikos.multi_user import __all__

        self.assertIn("MultiUserFilesystemBackend", __all__)
        self.assertIn("MultiUserXandikosApp", __all__)
        self.assertIn("InvalidUsernameError", __all__)
        self.assertIn("validate_username", __all__)
        self.assertIn("add_parser", __all__)
        self.assertIn("main", __all__)

    def test_multi_user_backend_importable(self):
        """Test that MultiUserFilesystemBackend can be imported directly."""
        from xandikos.multi_user import MultiUserFilesystemBackend

        self.assertIsNotNone(MultiUserFilesystemBackend)

    def test_add_parser_importable(self):
        """Test that add_parser can be imported directly."""
        from xandikos.multi_user import add_parser

        self.assertIsNotNone(add_parser)
        self.assertTrue(callable(add_parser))

    def test_main_importable(self):
        """Test that main can be imported directly."""
        from xandikos.multi_user import main

        self.assertIsNotNone(main)
        self.assertTrue(callable(main))



# Xandikos
# Copyright (C) 2025-2026 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Shared fixtures for Xandikos benchmarks.

These benchmarks are designed to run against different xandikos versions
(v0.3.0, v0.3.4, v0.3.5, HEAD) to detect performance regressions.

Usage:
    # Run benchmarks on current checkout:
    pytest benchmarks/ --benchmark-enable

    # Save results for later comparison:
    pytest benchmarks/ --benchmark-enable --benchmark-save=v0.3.5

    # Compare saved results:
    pytest-benchmark compare v0.3.4 v0.3.5 master --sort=fullname
"""

import inspect
import os
import shutil
import tempfile
from datetime import datetime, timedelta, timezone

import pytest

from xandikos.icalendar import ICalendarFile
from xandikos.store.git import BareGitStore, TreeGitStore
from xandikos.store.memory import MemoryStore


# Number of items in collections used for benchmarks.
SMALL_COLLECTION = 50
LARGE_COLLECTION = 500


def _make_vcalendar(i: int, base_date: datetime) -> bytes:
    """Generate a VCALENDAR with a single VEVENT."""
    event_date = base_date + timedelta(days=i)
    end_date = event_date + timedelta(hours=1)
    return (
        b"BEGIN:VCALENDAR\r\n"
        b"VERSION:2.0\r\n"
        b"PRODID:-//Xandikos Bench//EN\r\n"
        b"BEGIN:VEVENT\r\n"
        + f"UID:bench-event-{i}@example.com\r\n".encode()
        + f"DTSTART:{event_date.strftime('%Y%m%dT%H%M%SZ')}\r\n".encode()
        + f"DTEND:{end_date.strftime('%Y%m%dT%H%M%SZ')}\r\n".encode()
        + f"SUMMARY:Benchmark Event {i}\r\n".encode()
        + b"END:VEVENT\r\n"
        b"END:VCALENDAR\r\n"
    )


def _import_one(store, name, content_type, data):
    """Call import_one with the right kwargs for the current xandikos version.

    v0.3.0-v0.3.4 use ``author=``, v0.3.5+ use ``remote_user=``.
    """
    sig = inspect.signature(store.import_one)
    if "remote_user" in sig.parameters:
        return store.import_one(name, content_type, data)
    else:
        return store.import_one(name, content_type, data)


def _populate_store(store, n):
    """Add *n* calendar events to *store*, return {name: etag} map."""
    store.load_extra_file_handler(ICalendarFile)
    base_date = datetime(2025, 1, 1, 10, 0, 0, tzinfo=timezone.utc)
    etags = {}
    for i in range(n):
        name = f"event-{i}.ics"
        (name, etag) = _import_one(
            store, name, "text/calendar", [_make_vcalendar(i, base_date)]
        )
        etags[name] = etag
    return etags


# -- BareGitStore fixtures ---------------------------------------------------


@pytest.fixture(scope="session")
def bare_store_small():
    store = BareGitStore.create_memory()
    etags = _populate_store(store, SMALL_COLLECTION)
    return store, etags


@pytest.fixture(scope="session")
def bare_store_large():
    store = BareGitStore.create_memory()
    etags = _populate_store(store, LARGE_COLLECTION)
    return store, etags


# -- TreeGitStore fixtures ----------------------------------------------------


@pytest.fixture(scope="session")
def _tree_store_dir():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


@pytest.fixture(scope="session")
def tree_store_small(_tree_store_dir):
    store = TreeGitStore.create(os.path.join(_tree_store_dir, "small"))
    etags = _populate_store(store, SMALL_COLLECTION)
    return store, etags


@pytest.fixture(scope="session")
def tree_store_large(_tree_store_dir):
    store = TreeGitStore.create(os.path.join(_tree_store_dir, "large"))
    etags = _populate_store(store, LARGE_COLLECTION)
    return store, etags


# -- MemoryStore fixtures -----------------------------------------------------


@pytest.fixture(scope="session")
def memory_store_small():
    store = MemoryStore()
    etags = _populate_store(store, SMALL_COLLECTION)
    return store, etags


@pytest.fixture(scope="session")
def memory_store_large():
    store = MemoryStore()
    etags = _populate_store(store, LARGE_COLLECTION)
    return store, etags


# -- helpers available to test modules ----------------------------------------


def has_get_etag():
    """True if the installed xandikos has the direct get_etag() method."""
    return hasattr(BareGitStore, "get_etag") and not isinstance(
        BareGitStore.__dict__.get("get_etag"), property
    )


def has_get_file_meta():
    """True if the installed xandikos has get_file_meta()."""
    from xandikos.store import Store

    return hasattr(Store, "get_file_meta")



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Git store."""

import configparser
import errno
import functools
from logging import getLogger
import os
import shutil
import stat
import uuid
from io import BytesIO, StringIO
from typing import cast
from collections.abc import Iterable

import dulwich.repo
from dulwich.file import FileLocked
from dulwich.index import IndexEntry, index_entry_from_stat, locked_index
from dulwich.objects import Blob, Commit, Tree
from dulwich.porcelain import get_user_identity

from . import (
    DEFAULT_MIME_TYPE,
    MIMETYPES,
    VALID_STORE_TYPES,
    DuplicateUidError,
    InvalidCTag,
    InvalidETag,
    InvalidFileContents,
    LockedError,
    NoSuchItem,
    NotStoreError,
    OutOfSpaceError,
    Store,
    open_by_content_type,
    open_by_extension,
)
from .config import CONFIG_FILENAME
from .config import CollectionMetadata, FileBasedCollectionMetadata, is_metadata_file
from .index import MemoryIndex

DEFAULT_ENCODING = "utf-8"
DEFAULT_FILE_CACHE_SIZE = 1024


logger = getLogger("xandikos")


class RepoCollectionMetadata(CollectionMetadata):
    def __init__(self, repo) -> None:
        self._repo = repo
        self._cached_config = None
        self._cached_config_stat: tuple[float, int] | None = None

    def _get_config(self):
        """Return the repo config, re-reading only if the file changed."""
        if self._cached_config is not None:
            config_path = getattr(self._cached_config, "path", None)
            if config_path is not None:
                try:
                    st = os.stat(config_path)
                    current_stat = (st.st_mtime, st.st_size)
                except FileNotFoundError:
                    current_stat = None
            else:
                current_stat = None
            if self._cached_config_stat == current_stat:
                return self._cached_config

        config = self._repo.get_config()
        config_path = getattr(config, "path", None)
        if config_path is not None:
            try:
                st = os.stat(config_path)
                self._cached_config_stat = (st.st_mtime, st.st_size)
            except FileNotFoundError:
                self._cached_config_stat = None
        else:
            self._cached_config_stat = None
        self._cached_config = config
        return config

    @classmethod
    def present(cls, repo):
        config = repo.get_config()
        return config.has_section((b"xandikos",))

    def get_source_url(self):
        config = self._get_config()
        url = config.get(b"xandikos", b"source")
        if not url:
            raise KeyError
        return url.decode(DEFAULT_ENCODING)

    def set_source_url(self, url):
        config = self._get_config()
        if url is not None:
            config.set(b"xandikos", b"source", url.encode(DEFAULT_ENCODING))
        else:
            config.remove(b"xandikos", b"source")
        self._write_config(config)

    def get_color(self):
        config = self._get_config()
        color = config.get(b"xandikos", b"color")
        if color == b"":
            raise KeyError
        return color.decode(DEFAULT_ENCODING)

    def set_color(self, color):
        config = self._get_config()
        if color is not None:
            config.set(b"xandikos", b"color", color.encode(DEFAULT_ENCODING))
        else:
            config.remove(b"xandikos", b"color")
        self._write_config(config)

    def _write_config(self, config):
        f = BytesIO()
        config.write_to_file(f)
        self._repo._put_named_file("config", f.getvalue())
        # Update cache after write
        self._cached_config = config
        config_path = getattr(config, "path", None)
        if config_path is not None:
            try:
                st = os.stat(config_path)
                self._cached_config_stat = (st.st_mtime, st.st_size)
            except FileNotFoundError:
                self._cached_config_stat = None
        else:
            self._cached_config_stat = None

    def get_displayname(self):
        config = self._get_config()
        displayname = config.get(b"xandikos", b"displayname")
        if displayname == b"":
            raise KeyError
        return displayname.decode(DEFAULT_ENCODING)

    def set_displayname(self, displayname):
        config = self._get_config()
        if displayname is not None:
            config.set(
                b"xandikos",
                b"displayname",
                displayname.encode(DEFAULT_ENCODING),
            )
        else:
            config.remove(b"xandikos", b"displayname")
        self._write_config(config)

    def get_description(self):
        desc = self._repo.get_description()
        if desc in (None, b""):
            raise KeyError
        return desc.decode(DEFAULT_ENCODING)

    def set_description(self, description):
        if description is not None:
            self._repo.set_description(description.encode(DEFAULT_ENCODING))
        else:
            self._repo.set_description(b"")

    def get_comment(self):
        config = self._get_config()
        comment = config.get(b"xandikos", b"comment")
        if comment == b"":
            raise KeyError
        return comment.decode(DEFAULT_ENCODING)

    def set_comment(self, comment):
        config = self._get_config()
        if comment is not None:
            config.set(b"xandikos", b"comment", comment.encode(DEFAULT_ENCODING))
        else:
            config.remove(b"xandikos", b"comment")
        self._write_config(config)

    def set_type(self, store_type):
        config = self._get_config()
        config.set(b"xandikos", b"type", store_type.encode(DEFAULT_ENCODING))
        self._write_config(config)

    def get_type(self):
        config = self._get_config()
        store_type = config.get(b"xandikos", b"type")
        store_type = store_type.decode(DEFAULT_ENCODING)
        if store_type not in VALID_STORE_TYPES:
            logger.warning("Invalid store type %s set for %r.", store_type, self._repo)
        return store_type

    def get_order(self):
        config = self._get_config()
        order = config.get(b"xandikos", b"calendar-order")
        if order == b"":
            raise KeyError
        return order.decode("utf-8")

    def set_order(self, order):
        config = self._get_config()
        if order is None:
            order = ""
        config.set(b"xandikos", b"calendar-order", order.encode("utf-8"))
        self._write_config(config)

    def get_refreshrate(self):
        config = self._get_config()
        refreshrate = config.get(b"xandikos", b"refreshrate")
        if refreshrate == b"":
            raise KeyError
        return refreshrate.decode(DEFAULT_ENCODING)

    def set_refreshrate(self, refreshrate):
        config = self._get_config()
        if refreshrate is not None:
            config.set(
                b"xandikos", b"refreshrate", refreshrate.encode(DEFAULT_ENCODING)
            )
        else:
            config.remove(b"xandikos", b"refreshrate")
        self._write_config(config)

    def get_timezone(self):
        config = self._get_config()
        timezone = config.get(b"xandikos", b"timezone")
        if timezone == b"":
            raise KeyError
        return timezone.decode(DEFAULT_ENCODING)

    def set_timezone(self, timezone):
        config = self._get_config()
        if timezone is not None:
            config.set(b"xandikos", b"timezone", timezone.encode(DEFAULT_ENCODING))
        else:
            config.remove(b"xandikos", b"timezone")
        self._write_config(config)


class GitStore(Store):
    """A Store backed by a Git Repository."""

    def __init__(
        self,
        repo,
        *,
        ref: bytes = b"HEAD",
        check_for_duplicate_uids=True,
        parsed_file_cache_size: int | None = None,
        **kwargs,
    ) -> None:
        super().__init__(MemoryIndex(), **kwargs)
        self.ref = repo.refs.follow(ref)[0][-1]
        self.repo = repo
        # Disable automatic garbage collection
        self.repo._autogc_disabled = True
        # Maps uids to (sha, fname)
        self._uid_to_fname: dict[str, tuple[bytes, str]] = {}
        self._check_for_duplicate_uids = check_for_duplicate_uids
        # Set of blob ids that have already been scanned
        self._fname_to_uid: dict[str, tuple[str, str]] = {}

        # Cache for guessed store type (when not set in git config)
        self._guessed_type: str | None = None
        self._guessed_type_ctag: str | None = None

        # Cache parsed files by blob SHA - avoids reparsing identical content
        if parsed_file_cache_size is None:
            parsed_file_cache_size = DEFAULT_FILE_CACHE_SIZE
        self._parsed_file_cache = functools.lru_cache(maxsize=parsed_file_cache_size)(
            self._parse_file_by_sha
        )

    def _parse_file_by_sha(self, sha: str, content_type: str | None, name: str):
        """Parse a file by its SHA, used for caching."""
        blob = self.repo.object_store[sha.encode("ascii")]
        if content_type is None:
            return open_by_extension(
                blob.chunked,
                name,
                extra_file_handlers=self.extra_file_handlers,
            )
        else:
            return open_by_content_type(
                blob.chunked,
                content_type,
                extra_file_handlers=self.extra_file_handlers,
            )

    def get_file(
        self, name: str, content_type: str | None = None, etag: str | None = None
    ):
        """Get file with caching based on blob SHA."""
        if etag is None:
            etag = self.get_etag(name)

        # Use cached parsing based on blob SHA
        return self._parsed_file_cache(etag, content_type, name)

    def get_etag(self, name: str) -> str:
        raise NotImplementedError(self.get_etag)

    def _import_one(
        self,
        name: str,
        data: Iterable[bytes],
        message: str,
    ):
        raise NotImplementedError(self._import_one)

    @property
    def config(self):
        if RepoCollectionMetadata.present(self.repo):
            return RepoCollectionMetadata(self.repo)
        else:
            cp = configparser.ConfigParser()
            try:
                cf = self._get_raw(CONFIG_FILENAME)
            except KeyError:
                pass
            else:
                if cf is not None:
                    cp.read_string(b"".join(cf).decode("utf-8"))

            def save_config(cp, message):
                f = StringIO()
                cp.write(f)
                self._import_one(
                    CONFIG_FILENAME, [f.getvalue().encode("utf-8")], message
                )

            return FileBasedCollectionMetadata(cp, save=save_config)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.repo!r}, ref={self.ref!r})"

    @property
    def path(self):
        return self.repo.path

    def _check_duplicate(self, uid, name, replace_etag):
        if uid is not None and self._check_for_duplicate_uids:
            self._scan_uids()
            try:
                (existing_name, _) = self._uid_to_fname[uid]
            except KeyError:
                pass
            else:
                if existing_name != name:
                    raise DuplicateUidError(uid, existing_name, name)

        try:
            etag = self.get_etag(name)
        except KeyError:
            etag = None
        if replace_etag is not None and etag != replace_etag:
            raise InvalidETag(name, etag, replace_etag)
        return etag

    def import_one(
        self,
        name: str,
        content_type: str,
        data: Iterable[bytes],
        message: str | None = None,
        replace_etag: str | None = None,
        remote_user: str | None = None,
        requester: str | None = None,
    ) -> tuple[str, str]:
        """Import a single object.

        Args:
          name: name of the object
          content_type: Content type
          data: serialized object as list of bytes
          message: Commit message
          replace_etag: optional etag of object to replace
          remote_user: Optional user name of the actor
          requester: Optional User-Agent or client information
        Raises:
          InvalidETag: when the name already exists but with different etag
          DuplicateUidError: when the uid already exists
        Returns: etag
        """
        if content_type is None:
            fi = open_by_extension(data, name, self.extra_file_handlers)
        else:
            fi = open_by_content_type(data, content_type, self.extra_file_handlers)
        if name is None:
            name = str(uuid.uuid4())
            extension = MIMETYPES.guess_extension(content_type)
            if extension is not None:
                name += extension
        fi.validate()
        try:
            uid = fi.get_uid()
        except (KeyError, NotImplementedError):
            uid = None
        self._check_duplicate(uid, name, replace_etag)
        if message is None:
            try:
                old_fi = self.get_file(name, content_type, replace_etag)
            except KeyError:
                old_fi = None
            message = "\n".join(fi.describe_delta(name, old_fi))

        if remote_user is not None or requester is not None:
            message += "\n"
            if remote_user is not None:
                message += f"\nRemote-User: {remote_user}"
            if requester is not None:
                message += f"\nRequester: {requester}"

        etag = self._import_one(name, fi.normalized(), message)
        etag_str = etag.decode("ascii")
        self._index_file(name, etag_str, fi)
        return (name, etag_str)

    def _get_raw(self, name, etag=None):
        """Get the raw contents of an object.

        Args:
          name: Name of the item
          etag: Optional etag
        Returns: raw contents as chunks
        """
        if etag is None:
            etag = self.get_etag(name)
        blob = self.repo.object_store[etag.encode("ascii")]
        return blob.chunked

    def _scan_uids(self):
        removed = set(self._fname_to_uid.keys())
        for name, mode, sha in self._iterblobs():
            etag = sha.decode("ascii")
            if name in removed:
                removed.remove(name)
            if name in self._fname_to_uid and self._fname_to_uid[name][0] == etag:
                continue
            blob = self.repo.object_store[sha]
            fi = open_by_extension(blob.chunked, name, self.extra_file_handlers)
            try:
                uid = fi.get_uid()
            except KeyError:
                logger.warning("No UID found in file %s", name)
                uid = None
            except InvalidFileContents as e:
                logger.warning("Unable to parse file %s: %s", name, e)
                uid = None
            except NotImplementedError:
                # This file type doesn't support UIDs
                uid = None
            self._fname_to_uid[name] = (etag, uid)
            if uid is not None:
                self._uid_to_fname[uid] = (name, etag)
        for name in removed:
            (unused_etag, uid) = self._fname_to_uid[name]
            if uid is not None:
                del self._uid_to_fname[uid]
            del self._fname_to_uid[name]

    def _iterblobs(self, ctag=None):
        raise NotImplementedError(self._iterblobs)

    def iter_with_etag(self, ctag=None):
        """Iterate over all items in the store with etag.

        Args:
          ctag: Ctag to iterate for
        Returns: iterator over (name, content_type, etag) tuples
        """
        for name, mode, sha in self._iterblobs(ctag):
            (mime_type, _) = MIMETYPES.guess_type(name)
            if mime_type is None:
                mime_type = DEFAULT_MIME_TYPE
            yield (name, mime_type, sha.decode("ascii"))

    @classmethod
    def create(cls, path):
        """Create a new store backed by a Git repository on disk.

        Returns: A `GitStore`
        """
        raise NotImplementedError(cls.create)

    @classmethod
    def open_from_path(cls, path, **kwargs):
        """Open a GitStore from a path.

        Args:
          path: Path
        Returns: A `GitStore`
        """
        try:
            repo = dulwich.repo.Repo(path)
            repo._autogc_disabled = True
            return cls.open(repo, **kwargs)
        except dulwich.repo.NotGitRepository:
            raise NotStoreError(path)

    @classmethod
    def open(cls, repo, **kwargs):
        """Open a GitStore given a Repo object.

        Args:
          repo: A Dulwich `Repo`
        Returns: A `GitStore`
        """
        if repo.has_index():
            return TreeGitStore(repo, **kwargs)
        else:
            return BareGitStore(repo, **kwargs)

    def get_description(self):
        """Get extended description.

        Returns: repository description as string
        """
        try:
            return self.config.get_description()
        except KeyError:
            return None

    def set_description(self, description):
        """Set extended description.

        Args:
          description: repository description as string
        """
        self.config.set_description(description)

    def set_comment(self, comment):
        """Set comment.

        Args:
          comment: Comment
        """
        self.config.set_comment(comment)

    def get_comment(self):
        """Get comment.

        Returns: Comment
        """
        try:
            return self.config.get_comment()
        except KeyError:
            return None

    def get_color(self):
        """Get color.

        Returns: A Color code, or None
        """
        try:
            return self.config.get_color()
        except KeyError:
            return None

    def set_color(self, color):
        """Set the color code for this store."""
        self.config.set_color(color)

    def get_source_url(self):
        """Get source URL."""
        try:
            return self.config.get_source_url()
        except KeyError:
            return None

    def set_source_url(self, url):
        """Set the source URL."""
        self.config.set_source_url(url)

    def get_displayname(self):
        """Get display name.

        Returns: The display name, or None if not set
        """
        try:
            return self.config.get_displayname()
        except KeyError:
            return None

    def set_displayname(self, displayname):
        """Set the display name.

        Args:
          displayname: New display name
        """
        self.config.set_displayname(displayname)

    def set_type(self, store_type):
        """Set store type.

        Args:
          store_type: New store type (one of VALID_STORE_TYPES)
        """
        self.config.set_type(store_type)

    def get_type(self):
        """Get store type.

        This looks in git config first, then falls back to guessing.
        The guessed result is cached and invalidated when the ctag changes.
        """
        try:
            return self.config.get_type()
        except KeyError:
            ctag = self.get_ctag()
            if self._guessed_type is not None and self._guessed_type_ctag == ctag:
                return self._guessed_type
            store_type = super().get_type()
            self._guessed_type = store_type
            self._guessed_type_ctag = ctag
            return store_type

    def iter_changes(self, old_ctag, new_ctag):
        """Get changes between two versions of this store.

        Args:
          old_ctag: Old ctag (None for empty Store)
          new_ctag: New ctag
        Returns: Iterator over (name, content_type, old_etag, new_etag)
        """
        if old_ctag is None:
            t = Tree()
            self.repo.object_store.add_object(t)
            old_ctag = t.id.decode("ascii")
        previous = {
            name: (content_type, etag)
            for (name, content_type, etag) in self.iter_with_etag(old_ctag)
        }
        for name, new_content_type, new_etag in self.iter_with_etag(new_ctag):
            try:
                (old_content_type, old_etag) = previous[name]
            except KeyError:
                old_etag = None
            else:
                assert old_content_type == new_content_type
            if old_etag != new_etag:
                yield (name, new_content_type, old_etag, new_etag)
            if old_etag is not None:
                del previous[name]
        for name, (old_content_type, old_etag) in previous.items():
            yield (name, old_content_type, old_etag, None)

    def destroy(self):
        """Destroy this store."""
        shutil.rmtree(self.path)


class BareGitStore(GitStore):
    """A Store backed by a bare git repository."""

    def __init__(self, repo, **kwargs) -> None:
        super().__init__(repo, **kwargs)
        self._cached_tree = None
        self._cached_ref_target = None

    def _get_current_tree(self):
        try:
            current_ref = self.repo.refs[self.ref]
        except KeyError:
            self._cached_tree = None
            self._cached_ref_target = None
            return Tree()
        if current_ref == self._cached_ref_target and self._cached_tree is not None:
            return self._cached_tree
        ref_object = self.repo[current_ref]
        if isinstance(ref_object, Tree):
            tree = ref_object
        else:
            tree = self.repo.object_store[ref_object.tree]
        self._cached_tree = tree
        self._cached_ref_target = current_ref
        return tree

    def get_etag(self, name):
        tree = self._get_current_tree()
        name = name.encode(DEFAULT_ENCODING)
        return tree[name][1].decode("ascii")

    def get_ctag(self):
        """Return the ctag for this store."""
        return self._get_current_tree().id.decode("ascii")

    def _iterblobs(self, ctag=None):
        if ctag is None:
            tree = self._get_current_tree()
        else:
            try:
                tree = self.repo.object_store[ctag.encode("ascii")]
            except KeyError as exc:
                raise InvalidCTag(ctag) from exc
        for name, mode, sha in tree.iteritems():
            name = name.decode(DEFAULT_ENCODING)
            if is_metadata_file(name):
                continue
            yield (name, mode, sha)

    @classmethod
    def create_memory(cls, **kwargs) -> "GitStore":
        """Create a new store backed by a memory repository.

        Returns: A `GitStore`
        """
        repo = dulwich.repo.MemoryRepo()
        return cls(repo, **kwargs)

    def _commit_tree(self, tree_id, message):
        """Create a commit for the given tree.

        Args:
            tree_id: Tree object ID
            message: Commit message (bytes)

        Returns:
            Commit SHA
        """
        import time

        c = Commit()
        c.tree = tree_id

        author = get_user_identity(self.repo.get_config_stack())
        timestamp = int(time.time())

        c.author = author
        c.committer = author
        c.author_time = timestamp
        c.commit_time = timestamp
        c.author_timezone = 0
        c.commit_timezone = 0
        c.encoding = b"UTF-8"
        c.message = message

        # Get parent commits
        try:
            c.parents = [self.repo.refs[self.ref]]
        except KeyError:
            c.parents = []

        # Add commit to object store
        self.repo.object_store.add_object(c)

        # Update ref
        self.repo.refs[self.ref] = c.id

        return c.id

    def _import_one(
        self,
        name: str,
        data: Iterable[bytes],
        message: str,
    ) -> bytes:
        """Import a single object.

        Args:
          name: Optional name of the object
          data: serialized object as bytes
          message: optional commit message
        Returns: etag
        """
        b = Blob()
        b.chunked = data
        tree = self._get_current_tree()
        old_tree_id = tree.id
        name_enc = name.encode(DEFAULT_ENCODING)
        tree[name_enc] = (0o644 | stat.S_IFREG, b.id)
        self.repo.object_store.add_objects([(tree, ""), (b, name_enc)])
        if tree.id != old_tree_id:
            self._commit_tree(tree.id, message.encode(DEFAULT_ENCODING))
        return b.id

    def delete_one(
        self, name, message=None, etag=None, remote_user=None, requester=None
    ):
        """Delete an item.

        Args:
          name: Filename to delete
          message: Commit message
          etag: Optional mandatory etag of object to remove
          remote_user: Optional user name of the actor
          requester: Optional User-Agent or client information
        Raises:
          NoSuchItem: when the item doesn't exist
          InvalidETag: If the specified ETag doesn't match the current
        """
        tree = self._get_current_tree()
        name_enc = name.encode(DEFAULT_ENCODING)
        try:
            current_sha = tree[name_enc][1]
        except KeyError as exc:
            raise NoSuchItem(name) from exc
        if etag is not None and current_sha != etag.encode("ascii"):
            raise InvalidETag(name, etag, current_sha.decode("ascii"))
        del tree[name_enc]
        self.repo.object_store.add_objects([(tree, "")])
        if message is None:
            fi = open_by_extension(
                self.repo.object_store[current_sha].chunked,
                name,
                self.extra_file_handlers,
            )
            message = "Delete " + fi.describe(name)
        if remote_user is not None or requester is not None:
            message += "\n"
            if remote_user is not None:
                message += f"\nRemote-User: {remote_user}"
            if requester is not None:
                message += f"\nRequester: {requester}"
        self._commit_tree(tree.id, message.encode(DEFAULT_ENCODING))

    @classmethod
    def create(cls, path):
        """Create a new store backed by a Git repository on disk.

        Returns: A `GitStore`
        """
        os.mkdir(path)
        repo = dulwich.repo.Repo.init_bare(path)
        repo._autogc_disabled = True
        return cls(repo)

    def subdirectories(self):
        """Returns subdirectories to probe for other stores.

        Returns: List of names
        """
        # Or perhaps just return all subdirectories but filter out
        # Git-owned ones?
        return []


class TreeGitStore(GitStore):
    """A Store that backs onto a treefull Git repository."""

    def __init__(self, repo, **kwargs) -> None:
        super().__init__(repo, **kwargs)
        self._cached_index = None
        self._cached_index_stat: tuple[int, int] | None = None
        self._cached_ctag: str | None = None

    def _open_index(self) -> tuple["dulwich.index.Index", str | None]:
        """Return a cached git index and ctag, re-reading only if the file changed.

        Returns: (index, ctag) tuple where ctag may be None if not yet computed.
        """
        index_path = self.repo.index_path()
        try:
            st = os.stat(index_path)
        except FileNotFoundError:
            self._cached_index = None
            self._cached_index_stat = None
            self._cached_ctag = None
            return self.repo.open_index(), None
        current_stat = (st.st_mtime_ns, st.st_size)
        if self._cached_index is not None and self._cached_index_stat == current_stat:
            return self._cached_index, self._cached_ctag
        index = self.repo.open_index()
        self._cached_index = index
        self._cached_index_stat = current_stat
        self._cached_ctag = None
        return index, None

    def _invalidate_index_cache(self):
        self._cached_index = None
        self._cached_index_stat = None
        self._cached_ctag = None

    @classmethod
    def create(cls, path, bare=True):
        """Create a new store backed by a Git repository on disk.

        Returns: A `GitStore`
        """
        os.mkdir(path)
        repo = dulwich.repo.Repo.init(path)
        repo._autogc_disabled = True
        return cls(repo)

    def get_etag(self, name):
        index, _ctag = self._open_index()
        name = name.encode(DEFAULT_ENCODING)
        return index[name].sha.decode("ascii")

    def _commit_tree(self, index, message):
        tree = index.commit(self.repo.object_store)
        return self.repo.get_worktree().commit(
            message=message,
            tree=tree,
        )

    def _import_one(
        self,
        name: str,
        data: Iterable[bytes],
        message: str,
    ) -> bytes:
        """Import a single object.

        Args:
          name: name of the object
          data: serialized object as list of bytes
          message: Commit message
        Returns: etag
        """
        try:
            with locked_index(self.repo.index_path()) as index:
                p = os.path.join(self.repo.path, name)
                # Create directory if it doesn't exist
                os.makedirs(os.path.dirname(p), exist_ok=True)
                with open(p, "wb") as f:
                    f.writelines(data)
                st = os.lstat(p)
                # Read the file back to create the blob since data iterator is consumed
                with open(p, "rb") as f:
                    blob = Blob.from_string(f.read())
                encoded_name = name.encode(DEFAULT_ENCODING)
                try:
                    entry = index[encoded_name]
                except KeyError:
                    entry = None
                if entry is None or (
                    isinstance(entry, IndexEntry) and blob.id != entry.sha
                ):
                    self.repo.object_store.add_object(blob)
                    index[encoded_name] = index_entry_from_stat(st, blob.id)
                    self._commit_tree(index, message.encode(DEFAULT_ENCODING))
        except FileLocked as exc:
            raise LockedError(name) from exc
        except OSError as exc:
            if exc.errno == errno.ENOSPC:
                raise OutOfSpaceError() from exc
            raise
        finally:
            self._invalidate_index_cache()
        return blob.id

    def delete_one(
        self,
        name: str,
        message: str | None = None,
        etag: str | None = None,
        remote_user: str | None = None,
        requester: str | None = None,
    ) -> None:
        """Delete an item.

        Args:
          name: Filename to delete
          message: Commit message
          etag: Optional mandatory etag of object to remove
          remote_user: Optional user name of the actor
          requester: Optional User-Agent or client information
        Raise:
          NoSuchItem: when the item doesn't exist
          InvalidETag: If the specified ETag doesn't match the current
        """
        p = os.path.join(self.repo.path, name)
        try:
            with open(p, "rb") as f:
                # cast to Blob, because on dulwich < 0.24.0 Blob.from_string returns ShaFile
                current_blob = cast(Blob, Blob.from_string(f.read()))
        except FileNotFoundError as exc:
            raise NoSuchItem(name) from exc
        except IsADirectoryError as exc:
            raise NoSuchItem(name) from exc
        if message is None:
            fi = open_by_extension(current_blob.chunked, name, self.extra_file_handlers)
            message = "Delete " + fi.describe(name)
        if remote_user is not None or requester is not None:
            message += "\n"
            if remote_user is not None:
                message += f"\nRemote-User: {remote_user}"
            if requester is not None:
                message += f"\nRequester: {requester}"
        if etag is not None:
            with open(p, "rb") as f:
                current_etag = current_blob.id
            if etag.encode("ascii") != current_etag:
                raise InvalidETag(name, etag, current_etag.decode("ascii"))
        try:
            with locked_index(self.repo.index_path()) as index:
                os.unlink(p)
                del index[name.encode(DEFAULT_ENCODING)]
                self._commit_tree(index, message.encode(DEFAULT_ENCODING))
        except FileLocked:
            raise LockedError(name)
        finally:
            self._invalidate_index_cache()

    def get_ctag(self) -> str:
        """Return the ctag for this store."""
        index, ctag = self._open_index()
        if ctag is not None:
            return ctag
        ctag = index.commit(self.repo.object_store).decode("ascii")
        self._cached_ctag = ctag
        return ctag

    def _iterblobs(self, ctag=None):
        """Iterate over all items in the store with etag.

        :yield: (name, etag) tuples
        """
        if ctag is not None:
            try:
                tree = self.repo.object_store[ctag.encode("ascii")]
            except KeyError as exc:
                raise InvalidCTag(ctag) from exc
            for name, mode, sha in tree.iteritems():
                name = name.decode(DEFAULT_ENCODING)
                if is_metadata_file(name):
                    continue
                yield (name, mode, sha)
        else:
            index, _ctag = self._open_index()
            for name, sha, mode in index.iterobjects():
                name = name.decode(DEFAULT_ENCODING)
                if is_metadata_file(name):
                    continue
                yield (name, mode, sha)

    def subdirectories(self):
        """Returns subdirectories to probe for other stores.

        Returns: List of names
        """
        ret = []
        for name in os.listdir(self.path):
            if name == dulwich.repo.CONTROLDIR or is_metadata_file(name):
                continue
            p = os.path.join(self.path, name)
            if os.path.isdir(p):
                ret.append(name)
        return ret



# Xandikos
# Copyright (C) 2025 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.quota (RFC 4331)."""

import asyncio
import unittest

from xandikos import quota
from xandikos.webdav import ET


class QuotaAvailableBytesPropertyTests(unittest.TestCase):
    """Tests for QuotaAvailableBytesProperty (RFC 4331 Section 3)."""

    def test_property_name(self):
        """Test quota-available-bytes property name."""
        prop = quota.QuotaAvailableBytesProperty()
        self.assertEqual(prop.name, "{DAV:}quota-available-bytes")

    def test_property_attributes(self):
        """Test quota-available-bytes property attributes."""
        prop = quota.QuotaAvailableBytesProperty()
        self.assertFalse(prop.in_allprops)
        self.assertTrue(prop.live)
        self.assertIsNone(prop.resource_type)

    def test_get_value(self):
        """Test quota-available-bytes get_value method.

        RFC 4331 Section 3: Contains the number of bytes available
        to the user without exceeding quota.
        """

        async def run_test():
            prop = quota.QuotaAvailableBytesProperty()

            class MockResource:
                def get_quota_available_bytes(self):
                    return "1073741824"  # 1 GB

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/test", resource, el, {})

            self.assertEqual(el.text, "1073741824")

        asyncio.run(run_test())


class QuotaUsedBytesPropertyTests(unittest.TestCase):
    """Tests for QuotaUsedBytesProperty (RFC 4331 Section 4)."""

    def test_property_name(self):
        """Test quota-used-bytes property name."""
        prop = quota.QuotaUsedBytesProperty()
        self.assertEqual(prop.name, "{DAV:}quota-used-bytes")

    def test_property_attributes(self):
        """Test quota-used-bytes property attributes."""
        prop = quota.QuotaUsedBytesProperty()
        self.assertFalse(prop.in_allprops)
        self.assertTrue(prop.live)
        self.assertIsNone(prop.resource_type)

    def test_get_value(self):
        """Test quota-used-bytes get_value method.

        RFC 4331 Section 4: Contains the number of bytes used by
        the resource and all its children.
        """

        async def run_test():
            prop = quota.QuotaUsedBytesProperty()

            class MockResource:
                def get_quota_used_bytes(self):
                    return "524288000"  # 500 MB

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/test", resource, el, {})

            self.assertEqual(el.text, "524288000")

        asyncio.run(run_test())


if __name__ == "__main__":
    unittest.main()



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Scheduling.

See https://tools.ietf.org/html/rfc6638
"""

from xandikos import caldav, webdav
from xandikos.caldav import (
    SCHEDULE_INBOX_RESOURCE_TYPE,
    SCHEDULE_OUTBOX_RESOURCE_TYPE,
)

# Feature to advertise to indicate scheduling support.
FEATURE = "calendar-auto-schedule"

CALENDAR_USER_TYPE_INDIVIDUAL = "INDIVIDUAL"  # An individual
CALENDAR_USER_TYPE_GROUP = "GROUP"  # A group of individuals
CALENDAR_USER_TYPE_RESOURCE = "RESOURCE"  # A physical resource
CALENDAR_USER_TYPE_ROOM = "ROOM"  # A room resource
CALENDAR_USER_TYPE_UNKNOWN = "UNKNOWN"  # Otherwise not known

CALENDAR_USER_TYPES = (
    CALENDAR_USER_TYPE_INDIVIDUAL,
    CALENDAR_USER_TYPE_GROUP,
    CALENDAR_USER_TYPE_RESOURCE,
    CALENDAR_USER_TYPE_ROOM,
    CALENDAR_USER_TYPE_UNKNOWN,
)


class ScheduleInbox(webdav.Collection):
    resource_types = webdav.Collection.resource_types + [SCHEDULE_INBOX_RESOURCE_TYPE]

    def get_calendar_user_type(self):
        # Default, per section 2.4.2
        return CALENDAR_USER_TYPE_INDIVIDUAL

    def get_calendar_timezone(self):
        """Return calendar timezone.

        This should be an iCalendar object with exactly one
        VTIMEZONE component.
        """
        raise NotImplementedError(self.get_calendar_timezone)

    def set_calendar_timezone(self):
        """Set calendar timezone.

        This should be an iCalendar object with exactly one
        VTIMEZONE component.
        """
        raise NotImplementedError(self.set_calendar_timezone)

    def get_supported_calendar_components(self):
        """Return set of supported calendar components in this calendar.

        Returns: iterable over component names
        """
        raise NotImplementedError(self.get_supported_calendar_components)

    def get_supported_calendar_data_types(self):
        """Return supported calendar data types.

        Returns: iterable over (content_type, version) tuples
        """
        raise NotImplementedError(self.get_supported_calendar_data_types)

    def get_min_date_time(self):
        """Return minimum datetime property."""
        raise NotImplementedError(self.get_min_date_time)

    def get_max_date_time(self):
        """Return maximum datetime property."""
        raise NotImplementedError(self.get_max_date_time)

    def get_max_instances(self):
        """Return maximum number of instances."""
        raise NotImplementedError(self.get_max_instances)

    def get_max_attendees_per_instance(self):
        """Return maximum number of attendees per instance."""
        raise NotImplementedError(self.get_max_attendees_per_instance)

    def get_max_resource_size(self):
        """Return max resource size."""
        raise NotImplementedError(self.get_max_resource_size)

    def get_schedule_default_calendar_url(self):
        """Return default calendar URL.

        None indicates there is no default URL.
        """
        return None


class ScheduleOutbox(webdav.Collection):
    resource_types = webdav.Collection.resource_types + [SCHEDULE_OUTBOX_RESOURCE_TYPE]

    def get_supported_calendar_components(self):
        """Return set of supported calendar components in this calendar.

        Returns: iterable over component names
        """
        raise NotImplementedError(self.get_supported_calendar_components)

    def get_supported_calendar_data_types(self):
        """Return supported calendar data types.

        Returns: iterable over (content_type, version) tuples
        """
        raise NotImplementedError(self.get_supported_calendar_data_types)

    def get_max_resource_size(self):
        """Return max resource size."""
        raise NotImplementedError(self.get_max_resource_size)

    def get_min_date_time(self):
        """Return minimum datetime property."""
        raise NotImplementedError(self.get_min_date_time)

    def get_max_date_time(self):
        """Return maximum datetime property."""
        raise NotImplementedError(self.get_max_date_time)

    def get_max_attendees_per_instance(self):
        """Return maximum number of attendees per instance."""
        raise NotImplementedError(self.get_max_attendees_per_instance)


class ScheduleInboxURLProperty(webdav.Property):
    """Schedule-inbox-URL property.

    See https://tools.ietf.org/html/rfc6638, section 2.2
    """

    name = "{%s}schedule-inbox-URL" % caldav.NAMESPACE
    resource_type = webdav.PRINCIPAL_RESOURCE_TYPE
    in_allprops = True

    async def get_value(self, href, resource, el, environ):
        el.append(webdav.create_href(resource.get_schedule_inbox_url(), href))


class ScheduleOutboxURLProperty(webdav.Property):
    """Schedule-outbox-URL property.

    See https://tools.ietf.org/html/rfc6638, section 2.1
    """

    name = "{%s}schedule-outbox-URL" % caldav.NAMESPACE
    resource_type = webdav.PRINCIPAL_RESOURCE_TYPE
    in_allprops = True

    async def get_value(self, href, resource, el, environ):
        el.append(webdav.create_href(resource.get_schedule_outbox_url(), href))


class CalendarUserAddressSetProperty(webdav.Property):
    """calendar-user-address-set property.

    See https://tools.ietf.org/html/rfc6638, section 2.4.1
    """

    name = "{%s}calendar-user-address-set" % caldav.NAMESPACE
    resource_type = webdav.PRINCIPAL_RESOURCE_TYPE
    in_allprops = False

    async def get_value(self, base_href, resource, el, environ):
        for href in resource.get_calendar_user_address_set():
            el.append(webdav.create_href(href, base_href))


class ScheduleTagProperty(webdav.Property):
    """schedule-tag property.

    See https://tools.ietf.org/html/rfc6638, section 3.2.10
    """

    name = "{%s}schedule-tag" % caldav.NAMESPACE
    in_allprops = False

    def supported_on(self, resource):
        return resource.get_content_type() == "text/calendar"

    async def get_value(self, base_href, resource, el, environ):
        el.text = resource.get_schedule_tag()


class CalendarUserTypeProperty(webdav.Property):
    """calendar-user-type property.

    See https://tools.ietf.org/html/rfc6638, section 2.4.2
    """

    name = "{%s}calendar-user-type" % caldav.NAMESPACE
    resource_type = webdav.PRINCIPAL_RESOURCE_TYPE
    in_allprops = False

    async def get_value(self, href, resource, el, environ):
        el.text = resource.get_calendar_user_type()


class ScheduleDefaultCalendarURLProperty(webdav.Property):
    """schedule-default-calendar-URL property.

    See https://tools.ietf.org/html/rfc6638, section-9.2
    """

    name = "{%s}schedule-default-calendar-URL" % caldav.NAMESPACE
    resource_type = SCHEDULE_INBOX_RESOURCE_TYPE
    in_allprops = True

    async def get_value(self, href, resource, el, environ):
        url = resource.get_schedule_default_calendar_url()
        if url is not None:
            el.append(webdav.create_href(url, href))
